# coding: utf-8

import click
import signal
from deployv.base import errors
from deployv.base.worker import WorkerManager
from deployv.helpers import utils, configuration_helper


def signal_handler(signal_number, stack_frame):
    raise errors.GracefulExit(
        'Received a signal to terminate, stopping workers'
    )


signal.signal(signal.SIGTERM, signal_handler)
signal.signal(signal.SIGINT, signal_handler)


@click.command()
@click.option("-l", "--log_level", help="Log level to show", default='INFO')
@click.option("-h", "--log_file", help="Write log history to a file")
@click.option("-C", "--config", help="Additional .conf files.")
@click.option(
    "--worker_type", help="Parameter used to specify the type worker.")
def run(log_level, log_file, config, worker_type):
    utils.setup_deployv_logger(level=log_level, log_file=log_file)
    cfg = configuration_helper.DeployvConfig(worker_config=config)
    worker_tp = cfg.deployer.get('worker_type')
    module = __import__('deployv.messaging', fromlist=[str(worker_tp)])
    if not hasattr(module, worker_tp):
        raise ValueError
    msg_object = getattr(module, worker_tp)
    config_class = msg_object.CONFIG_CLASSES['file']
    keys = {'commit': 'Commit worker',
            'status': 'Status worker'}
    name = keys.get(worker_type, 'Deploy worker')
    # Pass the initialized manager to the class to avoid losing data when the worker reconnects.
    worker_manager = WorkerManager(name, cfg.deployer.get('workers'))
    worker = msg_object.factory(config_class(cfg, worker_type), name, worker_manager)
    while True:
        try:
            worker.run()
        except (KeyboardInterrupt, errors.GracefulExit):
            worker.signal_exit()
            break
        if not worker.should_reconnect():
            break
        # The old worker was killed and the connections it opened
        # have been closed by now, a new worker must be started
        worker = msg_object.factory(config_class(cfg, worker_type), name, worker_manager)



# coding: utf-8
import json
import logging
from deployv.helpers import configuration_helper
try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser


_logger = logging.getLogger('deployv')


class BaseHttpConfiguration:
    """ As the configuration for the sender can be from various sources this will be the base
    class for that, so there is no need to rewrite or change any code from the sender.
    Just need to be sure that the parameters are fully filled

    :param host: Hostname or ip where the odoo instance is running
    :param token: Token used to authenticate with the controller that returns the messages
    """
    def __init__(self, host, token):
        self.host = host
        self.controller = '{host}/messages_queue'.format(host=self.host)
        self.headers = {'content-type': 'application/json',
                        'http-worker-token': token}


class FileHttpConfiguration(BaseHttpConfiguration):
    """ Reads the configuration from a ini-style file and builds the base configuration

    :param config_object: File or Config parser object where the configuration is going to be read
    """
    def __init__(self, config_object, worker_type=False):
        if isinstance(config_object, ConfigParser):
            self.config = configuration_helper.DeployvConfig(worker_config=config_object)
        elif isinstance(config_object, configuration_helper.DeployvConfig):
            self.config = config_object
        else:
            raise ValueError('No valid config object provided')
        self._config_object = config_object
        self.wtype = worker_type
        host = self.config.http.get('orchest_host')
        token = self.config.http.get('token')
        super(FileHttpConfiguration, self).__init__(host, token)
        self.headers["is_status"] = json.dumps(worker_type == 'status')

    def get_result_object(self):
        return FileHttpConfiguration(self._config_object)



# coding: utf-8

import sys
import click
from pkg_resources import working_set
import deployv
from deployv.base import commandv
from deployv.helpers import utils, configuration_helper, json_helper
from deployv.instance import instancev
from deployv.base import postgresv
import subprocess
import os
import re
import logging

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class DeployvCLI(click.MultiCommand):

    _cmds = {}
    _build_warnings = []

    def __init__(self):
        super(DeployvCLI, self).__init__(
            invoke_without_command=True, no_args_is_help=True, callback=self._callback
        )
        addons = [[dist.key, dist.version]
                  for dist in working_set  # pylint: disable=E1133
                  if dist.key.startswith("deployv-addon-")]
        for addon in addons:
            addon_name = addon[0].replace('-', '_')
            addon_module = __import__(addon_name)
            command_name = addon_name.replace('deployv_addon_', '')
            try:
                command = getattr(addon_module, command_name)
            except AttributeError:
                self._build_warnings.append(
                    "Addon '{addon}' has no '{method}' method.".format(
                        addon=addon[0], method=command_name))
            else:
                try:
                    assert isinstance(command, click.core.Command)
                except AssertionError:
                    self._build_warnings.append(
                        "Error in addon '{addon}': Every command must be an instance"
                        " of <click.core.Command> class.'".format(addon=addon[0])
                    )
                else:
                    self.add_command(command.name, command)
        click.option("-l", "--log_level", help="Log level to show.", default='INFO')(self)
        click.option("-h", "--log_file", help="Write log history to a file.")(self)
        click.option("-C", "--config", help="Additional .conf files.")(self)

    def _callback(self, log_level, log_file, config):
        utils.setup_deployv_logger(level=log_level, log_file=log_file)
        logger.log(getattr(logging, log_level), 'Deployv version: %s', deployv.__version__)
        for warning in self._build_warnings:
            logger.debug(warning)
        self.config_parser = configuration_helper.DeployvConfig(worker_config=config)
        self.config = self.config_parser.worker_config

    def list_commands(self, ctx):
        return sorted(self._cmds)

    def get_command(self, ctx, name):
        return self._cmds[name] if name in self._cmds else None

    def command(self, **kwargs):
        def wrapper(function):
            doc = function.__doc__
            if doc and not kwargs.get('help'):
                kwargs['help'] = doc
            if doc and not kwargs.get('short_help'):
                kwargs['short_help'] = doc if len(doc) <= 50 else doc[:50] + '...'
                kwargs['short_help'] = doc

            def decorator():
                cmd = click.command(**kwargs)(function)
                return cmd
            self.add_command(function.__name__, decorator())
            return decorator
        return wrapper

    def add_command(self, name, function):
        self._cmds[name] = function


cli = DeployvCLI()  # pylint: disable=C0103


@cli.command()
@click.pass_context
@click.option("-f", "--config_file", help="Json file with the configuration",
              default=False, required=True)
@click.option("-z", "--backup-src", default=False, required=False,
              help=("Backup to be restored or directory where the backups are stored, if empty,"
                    " deployv will search for the backup_folder in the config json or in"
                    " deployv\'s config files if that parameter is not specified in the json"))
@click.option("-e", "--external", help=("Use external branches json file"
                                        " (generated with branchesv)"),
              default=False, required=False)
@click.option("-k", "--key_file", help="the key file to be used",
              default=False, required=False)
@click.option("--without-db", is_flag=True, default=False,
              help="If true, no database will be loaded inside the container. Default False")
@click.option("-d", "--database", help="Database name to be used",
              default=False, required=False)
@click.option("--without-demo", is_flag=True, default=False,
              help="If True, no database will be loaded demo data"
                   " inside the container. Default False")
@click.option("--skip-push", is_flag=True, default=False,
              help="If True, the images built during the create will not be pushed"
                   " to quay.io. Default: False")
def create(ctx, config_file, backup_src, external, key_file,
           without_db, database, without_demo, skip_push):
    """Creates an instance then loads a backup from backup_dir folder.
    """
    config_obj = ctx.parent.command.config_parser
    if ((external and not utils.validate_external_file(external)) or
            (key_file and not utils.validate_external_file(key_file))):
        sys.exit(1)
    config_dict = utils.merge_config(config_file, keys_file=key_file, branches=external)
    if not config_dict:
        sys.exit(1)
    config_dict.get("container_config").get("env_vars").update({"without_demo": without_demo})
    config_dict.get("instance").get("config").update({"db_name": database} if database else {})
    config_dict.get('instance').update({'restore_db': not without_db})
    if not config_dict.get('instance').get('config').get('db_owner'):
        config_dict.get('instance').get('config').update({
            'db_owner': config_obj.postgres.get('db_user')})
    if not config_dict.get('instance').get('config').get('db_owner_passwd'):
        config_dict.get('instance').get('config').update({
            'db_owner_passwd': config_obj.postgres.get('db_password')})
    config_dict.get('group_config').update({'backup_src': backup_src} if backup_src else {})
    config_dict.get('group_config').update({
        'command': 'deployvcmd.commandv.create',
        'create_container': True,
        'skip_push': skip_push
    })
    config_obj.generate_configuration(config_dict)
    command = commandv.CommandV(config_obj)
    command.create_trigger()


@cli.command()
@click.pass_context
@click.option("-f", "--config_file", help="Json file with the configuration",
              default=False, required=False)
@click.option("-z", "--backup-dir", default=False, required=False,
              help=("Directory where the backup will be stored, if empty, deployv will use the"
                    " backup_folder parameter in the config json (if specified) or"
                    " in deployv\'s config files"))
@click.option("-r", "--reason", help="Reason why the backup is being generated (optional)",
              default=False, required=False)
@click.option("--tmp-dir", help="Temp dir to use (optional)",
              default=False, required=False)
@click.option("-c", "--cformat", help="Compression format to be used (tar.bz2, tar.gz or tar)",
              default='bz2', required=False)
@click.option("-d", "--database", help="Database name",
              default=False, required=True)
@click.option("-n", "--container", help="Container name or id",
              default=False, required=False)
@click.option("-x", "--prefix", required=False,
              help="Prefix used for the database name and the backup database_name")
@click.option("--jobs", default=False, required=False, type=int,
              help=("The numbers of dump in parallel by dumping njobs tables simultaneously."
                    " This option reduces the time of the dump but it also increases"
                    " the load on the database server."))
def backupdb(ctx, config_file, backup_dir, reason, tmp_dir, database, cformat,
             container, prefix, jobs):
    """Generates a backup and saves it in backup_dir folder.
    """
    config_obj = ctx.parent.command.config_parser
    if (config_file and container) or (not config_file and not container):
        logger.error('You must use one and only one of the -n (container) and the -f (config_file)'
                     ' parameters. Use one or the other')
        sys.exit(1)
    if container:
        config_file = {'container_config': {'container_name': container}}
    config_dict = utils.merge_config(config_file)
    if not config_dict:
        sys.exit(1)
    group_config = {'cformat': cformat, 'reason': reason, 'prefix': prefix,
                    'backup_src': backup_dir, 'temp_folder': tmp_dir, 'jobs': jobs}
    for param, value in group_config.items():
        config_dict.get('group_config').update({param: value} if value else {})
    config_dict['instance']['config'].update({'db_name': database} if database else {})
    config_obj.generate_configuration(config_dict)
    command = commandv.CommandV(config_obj)
    command.backup_trigger()


@cli.command()
@click.pass_context
@click.option("-f", "--config_file", help="Json file with the configuration",
              default=False, required=True)
@click.option("-z", "--backup-src", default=False, required=False,
              help=("Backup to be restored or directory where the backups are stored, if empty,"
                    " deployv will search for the backup_folder in the config json or in"
                    " deployv\'s config files if that parameter is not specified in the json"))
@click.option("-d", "--database", help="Database name to be used",
              default=False, required=False)
@click.option("-n", "--container", help="Container name or id",
              default=False, required=False)
@click.option("-s", "--admin_pass", help="Admin password",
              default=False, required=False)
@click.option("-t", "--instance-type", type=click.Choice(instancev.INSTANCE_TYPES),
              help=("Specifies the type of instance. test, "
                    "develop, updates or production"))
@click.option("-x", "--prefix", required=False,
              help="Prefix used for the database name")
@click.option("--nginx-url", required=False,
              help=("Container's nginx url used to set the web url freeze parameter,"
                    " if empty the parameter will be deleted"))
@click.option("--jobs", default=False, required=False, type=int,
              help=("The numbers of jobs parallel for restore the database."
                    " This option can dramatically reduce the time to restore database"))
@click.option("--update-method", default='normal', required=False,
              type=click.Choice(['normal', 'cou']),
              help="Update the database using click-odoo-update")
def restore(ctx, config_file, backup_src, database, container, admin_pass,
            instance_type, prefix, nginx_url, jobs, update_method):
    """Restores a backup from a directory.
    """
    config_obj = ctx.parent.command.config_parser
    if config_file and container or not config_file and not container:
        logger.error(
            'You must use one, and only one of the -n or -f parameters.')
        sys.exit(1)
    if container:
        config_file = {'container_config': {'container_name': container}}
    config_dict = utils.merge_config(config_file)
    if not config_dict:
        sys.exit(1)
    config_dict['instance']['config'].update({'admin': admin_pass} if admin_pass else {})
    config_dict.get('instance').update(
        {'instance_type': instance_type} if instance_type else {})
    config_dict.get('container_config').update({'nginx_url': nginx_url} if nginx_url else {})
    if not config_dict.get('instance').get('config').get('db_owner'):
        config_dict.get('instance').get('config').update({
            'db_owner': config_obj.postgres.get('db_user')})
    if not config_dict.get('instance').get('config').get('db_owner_passwd'):
        config_dict.get('instance').get('config').update({
            'db_owner_passwd': config_obj.postgres.get('db_password')})
    if container and not prefix and not database:
        database = utils.generate_dbname(config_dict)
        if not database:
            logger.error(
                'Could not get the name of the database, '
                'please use the -x or -d parameter')
            sys.exit(1)
            config_dict['instance'].update({'restore_db': True})
    config_dict['instance']['config'].update({'db_name': database} if database else {})
    config_dict.get('group_config').update({'backup_src': backup_src} if backup_src else {})
    config_dict.get('group_config').update({'jobs': jobs} if jobs else {})
    config_dict.get('group_config').update({'prefix': prefix} if prefix else {})
    config_dict.get('group_config').update({
        'command': 'deployvcmd.commandv.restore',
        'update_method': update_method
    })
    config_obj.generate_configuration(config_dict)
    command = commandv.CommandV(config_obj)
    command.restore_trigger()


@cli.command()
@click.pass_context
@click.option("-f", "--config_file", help="Json file with the configuration",
              default=False, required=True)
@click.option("-d", "--database", help="Database name to be updated",
              default=False, required=False)
@click.option("--update-method", default='normal', required=False,
              type=click.Choice(['normal', 'cou']),
              help="Update the database using click-odoo-update")
def update(ctx, config_file, database, update_method):
    """Updates an existing instance, clone the branches to the specified ones and updates database.
    """
    config_obj = ctx.parent.command.config_parser
    config_dict = utils.merge_config(config_file)
    if not config_dict:
        sys.exit(1)
    config_dict['instance']['config'].update({'db_name': database} if database else {})
    config_dict.get('group_config').update({
        'command': 'deployvcmd.commandv.updatedb',
        'update_method': update_method
    })
    config_obj.generate_configuration(config_dict)
    command = commandv.CommandV(config_obj)
    command.create_trigger()


@cli.command()
@click.pass_context
@click.option("-n", "--container", help="Container name or id")
@click.option("-d", "--database", help="Database name to be used")
@click.option("-s", "--admin_password", help="Password used to log in as admin in odoo",
              required=False, default=False)
@click.option("--nginx-url", required=False,
              help=("Container's nginx url used to set the web url freeze parameter,"
                    " if empty the parameter will be deleted"))
def deactivate(ctx, container, database, admin_password, nginx_url):
    """Deactivates a specific database inside a container.
    """
    config_obj = ctx.parent.command.config_parser
    config_dict = {
        'container_config': {
            'container_name': container,
            'nginx_url': nginx_url
        },
        'instance': {
            'config': {
                'admin': admin_password,
                'db_name': database
            }
        },
        'group_config': {
            'deactivating_database': True
        }
    }
    config_obj.generate_configuration(config_dict)
    command = commandv.CommandV(config_obj)
    command.deactivate_trigger()


@cli.command()
@click.pass_context
@click.option("-x", "--prefix", required=True,
              help="Prefix used for the database name and the backup database_name")
@click.option("-p", "--store_path", help="Directory where the deactivated backups will be stored",
              default=False, required=True)
@click.option("-z", "--backup-src", default=False, required=False,
              help=("Backup that will be deactivated or folder with backups to deactivate the"
                    " latest backup for the customer (-x parameter), if empty, deployv will"
                    " search for the backup_folder parameter in deployv\'s config files"))
@click.option("-s", "--admin_password", help="Password used to log in as admin in odoo",
              required=True)
@click.option("-c", "--cformat", default='bz2', required=False,
              help=("Compression format to be used: bz2 for tar.bz2, gz for tar.gz,"
                    " or folder for a backup without compression"))
@click.option("--upload", default=False, required=False,
              help="URL to upload the deactivated backup")
@click.option("--remove", is_flag=True, default=False, required=False,
              help="If true, only remove backup after it have been"
                   " successfully uploaded to the server")
@click.option("--nginx-url", required=False,
              help=("Container's nginx url used to set the web url freeze parameter,"
                    " if empty the parameter will be deleted"))
@click.option("--no-fs", is_flag=True, default=False, required=False,
              help=("If true, does not include file store in backup"))
@click.option("--jobs", default=False, required=False, type=int,
              help=("The numbers of dump in parallel by dumping njobs tables simultaneously."
                    " This option reduces the time of the dump but it also increases"
                    " the load on the database server."))
@click.option("--migration", is_flag=True, default=False, required=False,
              help=("If true, the deactivated backup is generated in the format"
                    " required by odoo's migration scripts"))
@click.option("--image-name", help=(
    "The image where get the deactivate.jinja customer, in example format: "
    "quay.io/vauxoo/test:test-11.0-stable"))
def deactivate_backup(ctx, prefix, store_path, backup_src, admin_password,
                      cformat, upload, remove, nginx_url, no_fs, jobs, migration, image_name):
    config_obj = ctx.parent.command.config_parser
    db_config = config_obj.postgres.copy()
    db_config.update({'nginx_url': nginx_url, 'admin': admin_password})
    default_config = utils.load_default_config()
    default_config['group_config'].update({
        'jobs': jobs, 'no_fs': no_fs, 'store_path': store_path,
        'remove': remove, 'deactivating_backup': True, 'backup_src': backup_src,
        'cformat': cformat, 'prefix': prefix, 'upload': upload, 'migration': migration})
    default_config['instance']['config'].update(db_config)
    default_config['container_config'].update({'image_name': image_name})
    config_obj.generate_configuration(default_config)
    command = commandv.CommandV(config_obj)
    try:
        command.deactivate_backup_trigger()
    except Exception as error:
        msg = utils.get_error_message(error)
        logger.error("Error executing the `deactivate_backup` command: %s", msg)
        db_name = command._CommandV__config.postgres.get("deactivated_db_name")
        cmd = 'ls -a %s | grep -P "%s_*"' % (config_obj.store_path, db_name)
        command.clean_working_folders()
        postgres = postgresv.PostgresShell(utils.odoo2postgres(db_config))
        if db_name:
            postgres.drop(db_name, True)
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
            res = process.communicate()[0].strip().decode()
            for backup in res.split("\n"):
                if not backup:
                    continue
                backup_path = os.path.join(config_obj.store_path, backup)
                utils.clean_files(backup_path)


@cli.command()
@click.pass_context
@click.option("-f", "--config_file",
              help=("Config file where the parameters for the build will be taken from,"
                    " if the other parameters are specified, the config will be"
                    " updated accordingly. If not specified, a default config will be used."
                    " You have to use this parameter in order to build an image with more than"
                    " one base repo"))
@click.option("-u", "--repo", help="URL of the main github repository that will be used")
@click.option("-b", "--branch", help=("Branch of the repository that will be cloned."
                                      " If no branch is specified, then the version will be used"))
@click.option("-i", "--image-name",
              help=("Name of the docker image that will be used as a base to build the new image."
                    " Default: vauxoo/odoo-{version}-image"))
@click.option("-w", "--working-folder", default=False,
              help=("Name of the folder where the files required for the build will be stored"
                    "(default /tmp)"))
@click.option("-v", "--version",
              help=("Version of the oca dependencies that will be used in the image"
                    " e.g. 8.0, 9.0, etc."))
@click.option("--force", is_flag=True, default=False,
              help="Removes the working folder if it already exists before creating it again")
@click.option("-O", "--odoo-repo",
              help=("Specific odoo repository to be cloned instead of the default one, must be"
                    " in the format namespace/reponame#branch. Example: odoo/odoo#8.0"))
@click.option("-T", "--tag", default=False,
              help="Custom name for the new image, if not specified a name will be generated")
@click.option("--skip-push", is_flag=True, default=False,
              help="If True, the images built during the create will not be pushed"
                   " to quay.io. Default: False")
def build(ctx, config_file, repo, branch, image_name, working_folder,
          version, force, odoo_repo, tag, skip_push):
    config_obj = ctx.parent.command.config_parser
    config_dict = utils.merge_config(config_file or {})
    variables = {}
    if version:
        variables.update({'version': version})
    if image_name:
        variables.update({'base_image': image_name})
    if not config_dict:
        sys.exit(1)
    if repo and not version or version and not repo:
        logger.error('If you specify the repo or the version you have to specify'
                     ' the other one as well')
        sys.exit(1)
    if branch and not repo:
        logger.error('If you specify a branch you have to specify the repo')
        sys.exit(1)
    branch = branch or version
    if repo and branch:
        repo_name = repo.split('/')[-1].split('.')[0]
        repo_path = os.path.join('extra_addons', repo_name)
        utils.add_repo(branch, repo_name, repo_path, repo, config=config_dict,
                       extra={'main_repo': True})
    if odoo_repo:
        if not re.search('.*/.*#.*', odoo_repo):
            logger.error('The specified odoo repo does not have the correct format.'
                         ' Example format: odoo/odoo#8.0')
            sys.exit(1)
        parts = odoo_repo.split('#', 1)
        variables.update({'odoo_repo': parts[0], 'odoo_branch': parts[1]})
    if not config_dict.get('instance').get('config').get('db_owner'):
        config_dict.get('instance').get('config').update({
            'db_owner': config_obj.postgres.get('db_user')
        })
    if not config_dict.get('instance').get('config').get('db_owner_passwd'):
        config_dict.get('instance').get('config').update({
            'db_owner_passwd': config_obj.postgres.get('db_password')
        })
    config_dict.get('container_config').update({
        'build_image': True, 'working_folder': working_folder})
    config_dict.get('group_config').update({
        'version': version, 'force': force, 'tag': tag, 'skip_push': skip_push
    })
    config_dict.get("container_config").get("env_vars").update(variables)
    config_obj.generate_configuration(config_dict)
    command = commandv.CommandV(config_obj)
    command.build_image_trigger()


@cli.command(short_help="Creates the default configuration.")
@click.pass_context
def setup_config(ctx):
    """Creates the default configuration directories and files. These are
    located in `/etc/deployv` (system level) and `~/.config/deployv` (user
    level). You may have to use `sudo` to create the file at system level.
    """
    ctx.parent.command.config_parser.setup_config_dirs()


@cli.command()
@click.pass_context
@click.option('-f', '--config_file', help='Instance configuration file (json).')
@click.option('-n', '--container', help='Production container name.')
@click.option('-d', '--db_name', help='Production database name.')
@click.option('-u', '--db_user', help='Production database user.')
@click.option('-w', '--db_password', help='Production database password.')
@click.option('-h', '--db_host', help='Production database host.')
@click.option('-p', '--db_port', help='Production database port.')
@click.option('-z', '--backup_dir', help='Directory where the production backup will be saved.')
@click.option('-e', '--decompressed_backup_size',
              help=('Expected size of the backup after extracting it, to make sure there is'
                    ' enough space available before extracting it.'))
@click.option('-E', '--compressed_backup_size', default='',
              help=('Expected size of the backup that will be generated, to make sure there is'
                    ' enough space available before generating it. If not specified,'
                    ' the compressed backup size * 0.75 will be assumed'))
@click.option('-i', '--image', help='Stable image name to update the production instance.')
@click.option('-v', '--version',
              help='Odoo version to detect bin file. If empty, will be gotten from the env vars.')
@click.option('-t', '--status_checks', default=4,
              help="How many times it will try to check the status of Odoo (running/stopped).")
@click.option('-c', '--cformat', default='bz2',
              help='Compression format to be used (tar.bz2 or tar.gz).')
@click.option('-m', '--tmp-dir', help='Temp directory where the backup is going to be generated.')
@click.option('-x', '--prefix', help='Prefix for the backup name.')
@click.option('-r', '--reason', help='Reason for the backup name.', default='preupdate')
def update_production(ctx, config_file, container, db_name, db_user, db_password, db_host, db_port,
                      version, backup_dir, decompressed_backup_size, compressed_backup_size,
                      image, status_checks, cformat, tmp_dir, prefix, reason):
    """Updates a production instance.
    """
    config_obj = ctx.parent.command.config_parser
    config_dict = config_file and json_helper.load_json(config_file)
    if not config_dict:
        logger.error('Error loading the configuration dictionary.\n'
                     'You must specify the json configuration file path (-f).')
        sys.exit(1)
    config_dict['container_config'].update({'container_name': container} if container else {})
    config_dict['container_config'].update({'image_name': image} if image else {})
    if not config_dict['container_config'].get('container_name') and not (
            config_dict['instance']['customer_id'] and
            config_dict['instance']['task_id']):
        logger.error('You must specify the container name (-n). '
                     'Or you can set it in the instance config file (-f) '
                     '(instance > customer_id + task_id).')
        sys.exit(1)
    json_db_config = config_dict['instance'].get('config') or {}
    db_config = {
        'db_name': db_name or json_db_config.get('db_name'),
        'db_user': db_user or json_db_config.get('db_user'),
        'db_password': db_password or json_db_config.get('db_password'),
        'db_host': db_host or json_db_config.get('db_host'),
        'db_port': db_port or json_db_config.get('db_port'),
        'dbfilter': False,
    }
    if not all(db_config.values()):
        logger.error('Production database configuration is incomplete. '
                     'You must set name (-d), user (-u), password (-w), host (-h), and port (-p). '
                     'Or you can set it on the instance config file (-f) (instance > config).')
        sys.exit(1)
    config_dict.get('instance').get('config').update(db_config)
    if not config_dict['container_config'].get('image_name'):
        logger.error('You must specify the docker image name to be used (-i). '
                     'Or you can set it on the instance configuration file (-f) '
                     '(container_config > image_name).')
        sys.exit(1)
    group_config = {
        'decompressed_backup_size': decompressed_backup_size,
        'compressed_backup_size': compressed_backup_size,
        'version': version,
        'status_checks': status_checks,
        'cformat': cformat,
        'reason': reason,
        'backup_src': backup_dir,
        'temp_folder': tmp_dir,
        'prefix': prefix,
        'command': 'deployvcmd.commandv.update_production'
    }
    if not config_dict.get('group_config'):
        config_dict.update({'group_config': {}})
    for param, value in group_config.items():
        config_dict['group_config'].update({param: value} if value else {})
    if (not config_dict['group_config'].get('decompressed_backup_size') or not
            utils.parse_backup_size(config_dict['group_config'].get('decompressed_backup_size'))):
        logger.error('You must specify a valid expected size for the backup (-e). This to ensure '
                     'the space is available before generating the backup. For example: 50GB.')
        sys.exit(1)
    config_obj.generate_configuration(config_dict, load_defaults=False)
    command = commandv.CommandV(config_obj)
    command.update_prod_trigger()



import csv
import re

TYPES = ['L', 'N', 'A', 'C', 'E']   # literal, number, alpha, alphanumeric, escaped literal


def get_args_pattern(pattern):
    r"""Return format string and parameters from a received pattern.

    Examples:
    - Received pattern: "V\A111FYAAAZZ1" , Result: "VA%sFY%s%s%s", "[1, 3], [2, 3], [3, 2], [1, 1]"
    - Received pattern: "(+52)111-1111111" , Result: "(+52)%s-%s", "[1, 3], [1, 7]"

    Note: this function is somewhat tricky"""   # noqa: W605
    d_mapping = {'1': 'N', 'A': 'A', 'Z': 'C', '\\': 'E'}   # to get proper type value
    arg_list = ''
    arg_str = ''
    prev_ch_type = '*'
    qty = 0
    str_val = ''
    for ch in pattern:
        ch_type = d_mapping.get(ch, 'L')
        if ch_type == prev_ch_type:
            if ch_type in ['N', 'A', 'C']:
                qty += 1
            else:
                str_val += ch
        else:
            if prev_ch_type == 'E':
                ch_type = 'L'
            elif ch_type == 'E':
                pass
            elif prev_ch_type in ['N', 'A', 'C'] and qty:
                arg_str += '%s'
                arg_list += f"[{TYPES.index(prev_ch_type)}, {qty}], "
                str_val = ''
            else:
                if prev_ch_type and str_val:
                    arg_str += f'{str_val}'
                    str_val = ''
                # str_val += ch
            if ch_type == 'L':
                str_val += ch
            elif ch_type in ['N', 'A', 'C']:
                qty = 1
            prev_ch_type = ch_type
        last_ch = ch
    if prev_ch_type == 'E':
        str_val += last_ch
    elif prev_ch_type in ['N', 'A', 'C'] and qty:
        arg_str += '%s'
        arg_list += f'[{TYPES.index(prev_ch_type)}, {qty}]'
    else:
        arg_str += f'{str_val}'
    arg_list = arg_list.strip()
    if arg_list[-1] == ',':
        arg_list = arg_list[:-1]
    return arg_str, arg_list


def get_masking_function(table_name, field_name, field_type, argument):
    """Return SQL query for apply anonymize operation in a field of table

    Examples:
    - Received: "ir_mail_server", "smtp_pass", "alphanumeric", "16"
      Result: SECURITY LABEL FOR anon ON COLUMN ir_mail_server.smtp_pass
              IS 'MASKED WITH FUNCTION anon.random_string(16)';
    - Received: "res_bank", "name", "alpha", "10"
      Result: SECURITY LABEL FOR anon ON COLUMN res_bank.name
              IS 'MASKED WITH FUNCTION vx.random_alpha(10)';
    - Received: "res_partner", "city", "city", None
      Result: SECURITY LABEL FOR anon ON COLUMN res_partner.city
              IS 'MASKED WITH FUNCTION anon.fake_city()'; """
    d_mapping = {'firstname': "FUNCTION anon.fake_first_name()'; ",
                 'lastname': "FUNCTION anon.fake_last_name()'; ",
                 'name': "FUNCTION vx.random_name()'; ",
                 'email': "FUNCTION anon.fake_email()'; ",
                 'country': "FUNCTION anon.fake_country()'; ",
                 'city': "FUNCTION anon.fake_city()'; ",
                 'company': "FUNCTION anon.fake_company()'; "}
    if not table_name or not field_name or not field_type:
        return ''
    field_type = field_type.lower()
    regex_length = re.compile(r'(\d+)')
    regex_range = re.compile(r'(\d+)-(\d+)')
    query = f"SECURITY LABEL FOR anon ON COLUMN {table_name}.{field_name} IS 'MASKED WITH "
    if field_type in ['constantnum', 'constantstr', 'digit', 'alpha', 'alphanumeric', 'pattern'] \
            and not argument:
        return ''
    if field_type == 'constantnum':
        query += f"VALUE {argument}'; "
    elif field_type == 'constantstr':
        query += f"VALUE ''{argument}'' '; "
    elif field_type == 'digit':
        m_obj = regex_range.match(argument)
        if m_obj and m_obj.group(1) and m_obj.group(2):
            num1 = m_obj.group(1)
            num2 = m_obj.group(2)
        else:
            m_obj = regex_length.match(argument)
            if m_obj and m_obj.group(1):
                qty_digits = int(m_obj.group(1))
                num1 = '1' + ('0' * (qty_digits-1) if (qty_digits-1) else '')
                num2 = '9' * qty_digits
            else:
                return ''
        query += f"FUNCTION anon.random_int_between({num1}, {num2})'; "
    elif field_type == 'alpha':
        m_obj = regex_length.match(argument)
        if m_obj and m_obj.group(1):
            query += f"FUNCTION vx.random_alpha({m_obj.group(1)})'; "
        else:
            query = ''
    elif field_type == 'alphanumeric':
        m_obj = regex_length.match(argument)
        if m_obj and m_obj.group(1):
            query += f"FUNCTION anon.random_string({m_obj.group(1)})'; "
        else:
            query = ''
    elif field_type == 'pattern':
        format_str, format_arg_list = get_args_pattern(argument)
        query += f"FUNCTION vx.pattern(''{format_str}'', ARRAY [{format_arg_list}])'; "
    else:
        query += d_mapping.get(field_type, '')
    return query


def get_queries_from_operations_file(filename):
    """Read anonymize operations from a file and return equivalent sql queries"""
    regex = re.compile(r'table:(.*)', re.IGNORECASE)
    queries = []
    query = ''
    table_name = ''
    with open(filename) as fh:
        csv_reader = csv.reader(fh, delimiter=',')
        for row in csv_reader:
            if not row:
                continue
            res = regex.match(row[0])
            if res:
                if query and table_name:
                    query += f"SELECT anon.anonymize_table('{table_name}');"
                    queries.append(query)
                query = ''
                table_name = res.group(1)
                table_name = table_name.strip()
                continue
            field_name = row[0]
            field_type = row[1]
            if len(row) > 2:
                argument = row[2]
            else:
                argument = None
            if table_name:
                query += get_masking_function(table_name, field_name, field_type, argument)
        if query and table_name:
            query += f"SELECT anon.anonymize_table('{table_name}');"
            queries.append(query)
    return queries


def get_queries_vx_functions():
    return """CREATE OR REPLACE FUNCTION vx.random_alpha(k integer)
    RETURNS varchar AS
    $$
    DECLARE res varchar;
    BEGIN
    SELECT array_to_string(array(
    SELECT substr('ABCDEFGHIJKLMNOPQRSTUVWXYZ',((random()*(26-k)+1)::integer) ,1)
    FROM generate_series(1, k)
    ), '') INTO res;
    RETURN res;
    END
    $$
    LANGUAGE PLPGSQL
    VOLATILE
    RETURNS NULL ON NULL INPUT
    SECURITY INVOKER
    SET search_path='';
    CREATE OR REPLACE FUNCTION vx.random_digit(k integer)
    RETURNS varchar AS
    $$
    DECLARE res varchar;
    BEGIN
    SELECT array_to_string(array(
    SELECT substr('0123456789',((random()*(10-k)+1)::integer) ,1) FROM generate_series(1, k)
    ), '') INTO res;
    RETURN res;
    END
    $$
    LANGUAGE PLPGSQL
    VOLATILE
    RETURNS NULL ON NULL INPUT
    SECURITY INVOKER
    SET search_path='';
    CREATE OR REPLACE FUNCTION vx.random_name()
    RETURNS varchar AS
    $$
    DECLARE res varchar;
    BEGIN
    SELECT anon.fake_first_name() || ' ' || anon.fake_last_name() INTO res;
    RETURN res;
    END
    $$
    LANGUAGE PLPGSQL
    VOLATILE
    RETURNS NULL ON NULL INPUT
    SECURITY INVOKER
    SET search_path='';
    CREATE OR REPLACE FUNCTION vx.gen_random_str(ch_type integer, ch_qty integer)
    RETURNS varchar AS
    $$
    DECLARE res varchar;
    BEGIN
    if ch_type = 1 then
    res := vx.random_digit(ch_qty);
    elsif ch_type = 2 then
    res := vx.random_alpha(ch_qty);
    elsif ch_type = 3 then
    res := anon.random_string(ch_qty);
    else
    res := '';
    end if;
    RETURN res;
    END
    $$
    LANGUAGE PLPGSQL
    VOLATILE
    RETURNS NULL ON NULL INPUT
    SECURITY INVOKER
    SET search_path='';
    CREATE OR REPLACE FUNCTION vx.pattern(format_str varchar, params integer[][])
    RETURNS varchar AS
    $$
    DECLARE
    args varchar[];
    res varchar;
    BEGIN
    FOR i IN 1..array_length(params, 1)
    LOOP
    args := array_append(args, vx.gen_random_str(params[i][1], params[i][2]));
    END LOOP;
    SELECT format(format_str, VARIADIC args) INTO res;
    RETURN res;
    END
    $$
    LANGUAGE PLPGSQL
    VOLATILE
    RETURNS NULL ON NULL INPUT
    SECURITY INVOKER
    SET search_path='';"""



# coding: utf-8

from deployv.messaging.basemsg import BaseWorker
from deployv.base import errors
import logging


_logger = logging.getLogger('deployv')


class HttpWorker(BaseWorker):
    def __init__(self, configuration_object, sender_class, receiver_class,
                 worker_id):
        super(HttpWorker, self).__init__(configuration_object, sender_class, receiver_class,
                                         worker_id)

    def run(self):
        """ Worker method, runs the method that performs a request to the odoo controller
        and retrieves the messages
        """
        _logger.info("Worker %s - waiting for something to do", self._wid)
        try:
            self._receiver.run(self.callback)
        except errors.GracefulExit:
            # Catch the exception raised in the threads when they are being stopped
            pass

    def signal_exit(self):
        """ Exit when finished with current loop """
        self._receiver.stop()

    def kill(self):
        """ This kill immediately the process, should not be used """
        self._receiver.stop()
        self.terminate()
        self.join()

    def callback(self, body):
        message = self.check_message(body)
        if message:
            _logger.info('Received a message worker %s', self._wid)
            self.execute_rpc(message)



# coding: utf-8

import logging
import os
import shlex
import spur
import psycopg2
import psycopg2.extras
from psycopg2 import OperationalError, ProgrammingError
from deployv.base import errors
from deployv.helpers import utils

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class PostgresConnector:
    """A simple helper to perform a postgres connection and execute simple sql sentences.
    """

    __conn = None
    __cursor = None
    __allowed_keys = ['host', 'port', 'dbname', 'user', 'password']
    __str_conn = ''

    def __init__(self, config=None):
        if config is None:
            config = {}
            # TODO: generate proper connection str from config dict
            # str_conn = "dbname=postgres"  # Error message or connect?
        for key, value in config.items():
            if value is not None and key in self.__allowed_keys:
                self.__str_conn = '%s %s=%s' % (self.__str_conn, key, value)
            elif key == 'dbname' and value is None:
                self.__str_conn = '%s %s=%s' % (
                    self.__str_conn, 'dbname', 'postgres')
        logger.debug('Connection string: %s', self.__str_conn)
        try:
            logger.debug('Stabilishing connection with db server')
            self.__conn = psycopg2.connect(self.__str_conn)
            if config.get('isolation_level', False):
                self.__conn.set_isolation_level(0)
            self.__cursor = self.__conn.cursor(
                cursor_factory=psycopg2.extras.DictCursor)
        except (KeyboardInterrupt, errors.GracefulExit):
            raise
        except Exception as error:
            # TODO: log this then raise
            logger.debug(
                'Connection to database not established: %s', str(error))
            # Proper cleanup
            self.disconnect()
            raise

    def execute(self, sql_str, *args, **kwargs):
        """This method basically wraps *execute* cursor method from psycopg2.

        See `passing parameters to sql
        <http://initd.org/psycopg/docs/usage.html#passing-parameters-to-sql-queries>`_
        for more information.

        :param sql_str: Sql to be executed.
        :type sql_str: str
        :param args: Args, all will be passed to `cursor.execute
            <http://initd.org/psycopg/docs/cursor.html#cursor.execute>`_.
        :type args: tuple
        :returns: A dict if a select was performed. Bool if update/insert.
        :rtype: dict or bool
        """
        raw = kwargs.pop('raw', False)
        change = sql_str.lower().startswith((
            'insert into', 'update', 'create', 'delete', 'alter', 'reassign', 'grant', 'security'))
        try:
            logger.debug('SQL: %s', sql_str)
            self.__cursor.execute(sql_str, *args)
        except (KeyboardInterrupt, errors.GracefulExit):
            raise
        except Exception:
            self.__conn.rollback()
            raise
        if change:
            self.__conn.commit()
        if raw:
            return self.__cursor.fetchall()
        return utils.copy_list_dicts(self.__cursor) if not change else True

    def check_config(self):
        """Check if can connect to PostgreSQL server with the provided configuration.

        :returns: If succeeded or not.
        :rtype: bool
        """
        try:
            logger.debug('')
            res = self.execute("select version();")
            logger.debug('Postgres returned: %s', res)
        except (KeyboardInterrupt, errors.GracefulExit):
            raise
        except Exception as error:  # pylint: disable=W0703
            logger.error('PostgreSQL connection test failed %s',
                         str(error))
            return False
        return True

    def disconnect(self):
        if self.__cursor:
            logger.debug('disconnect: closing cursor')
            self.__cursor.close()
        if self.__conn:
            logger.debug('disconnect: closing connection')
            self.__conn.close()

    def __del__(self):
        self.disconnect()

    def __exit__(self, type, value, traceback):  # pylint: disable=W0622
        self.disconnect()

    def __enter__(self):
        return self


class PostgresShell:
    """This class is used to perform some Postgres operations that generally are done in the
    console.

    For example:

    * pg_dump
    * pg_restore
    * psql
    * dropdb

    All keys passed in the dictionary to the constructor must be in `libpq keys style
    <http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-PARAMKEYWORDS>`_.
    """

    __defaults = {'dbname': 'postgres'}
    __formats = {'host': '-h {host}', 'port': '-p {port}', 'user': '-U {user}'}

    def __init__(self, config):

        self.__config = config
        for key, value in self.__defaults.items():
            if not self.__config.get(key, False):
                self.__config.update({key: value})

    def __format_cmd(self, cmd, items):
        res = cmd
        for key, value in self.__formats.items():
            if key in items:
                res = '%s %s' % (res, value)
        return res

    def list_databases(self):
        sql = '''SELECT d.datname as "name",
                    u.usename as "owner",
                    pg_catalog.pg_encoding_to_char(d.encoding) as "encoding"
                    FROM pg_catalog.pg_database d
                    LEFT JOIN pg_catalog.pg_user u ON d.datdba = u.usesysid
                    ORDER BY 1;'''

        with PostgresConnector(self.__config) as db:
            res = db.execute(sql)
        return res

    def terminate_sessions(self, db_name, owner, owner_passwd):
        """operation. terminate sessions of users in database

        :param db_name: Database name to terminate session.
        :type db_name: str
        :param owner: User name with permission superuser.
        :type owner: str
        :param owner_passwd: Password of owner superuser.
        :type owner_passwd: str
        :returns: None
        """
        config = self.__config.copy()
        config.update({'dbname': "template1", 'user': owner,
                       'password': owner_passwd})
        sql = (("SELECT pg_terminate_backend(pid) from pg_stat_activity "
               "WHERE pid <> pg_backend_pid() AND datname = '{dbname}'")
               .format(dbname=db_name))
        logger.info('Terminating sessions in the db %s', db_name)
        try:
            with PostgresConnector(config) as db:
                db.execute(sql)
        except (ProgrammingError, OperationalError) as error:
            logger.error(("Failed to end the connections to the database %s"),
                         utils.get_error_message(error))

    def dump(self, database, destiny_file, jobs=False, exclude=False):
        """Dumps database using pg_dump in sql format.

        :param database: database to be dumped.
        :type database: str
        :param destiny_file: Filename with full path where the dump will be stored.
        :type destiny_file: str
        :returns: The full dump path and name with .sql extension if success.
        :rtype: str or None
        """
        logger.debug("Dumping database %s into %s folder",
                     database, destiny_file)
        if self.__config.get('password', False):
            os.environ['PGPASSWORD'] = self.__config.get('password')
        self.__config.update({'dbname': database})
        dump_cmd = 'pg_dump {dbname} -w -O -x -f {0}'
        if jobs:
            self.__config.update({'num': jobs})
            dump_cmd = 'pg_dump -w  -j {num} -F d {dbname} -f {0} -x'
        if exclude:
            for table in exclude.split(","):
                dump_cmd += " --exclude-table=%s " % (table)
        dump_cmd = self.__format_cmd(dump_cmd, self.__config)
        dump_cmd = dump_cmd.format(destiny_file, **self.__config)
        logger.debug('Dump command: %s', dump_cmd)
        shell = spur.LocalShell()
        try:
            shell.run(shlex.split(dump_cmd))
        except spur.results.RunProcessError as error:
            str_error = utils.decode(error.stderr_output)
            if 'does not exist' in str_error:
                msg = 'Database does not exists, check name and try again'
            else:
                msg = 'Could not dump database, error message: {0}'.format(str_error)
            logger.error(msg)
            raise errors.DumpError(msg)
        return destiny_file

    def drop(self, database_name, force=False):
        """Drops a database using *dropdb* command.

        :param database_name: Database name to be deleted.
        :type database_name: str
        :returns: If succeeded (True) or not (None).
        :rtype: bool or None
        """
        cnf = self.__config
        if force:
            self.terminate_sessions(database_name, cnf.get("owner"),
                                    cnf.get("owner_password"))
        cnf.update({'dbname': database_name})
        logger.debug('drop: %s', database_name)
        if self.__config.get('password', False):
            os.environ['PGPASSWORD'] = self.__config.get('password')
        dropdb_cmd = 'dropdb {dbname}'
        dropdb_cmd = self.__format_cmd(dropdb_cmd, cnf)
        dropdb_cmd = dropdb_cmd.format(**cnf)
        shell = spur.LocalShell()
        try:
            shell.run(shlex.split(dropdb_cmd))
        except spur.results.RunProcessError as error:
            str_error = utils.decode(error.stderr_output)
            if 'does not exist' in str_error:
                return True
            logger.error('Could not drop database, error message: %s', str_error)
            raise OperationalError(str_error)
        else:
            return True

    def create(self, database_name):
        """Executes a *createdb* command.

        :param database_name: Database name.
        :returns: The created database name.
        :rtype: str
        """
        cnf = self.__config
        cnf.update({'dbname': database_name})
        logger.debug("Creating database %s", database_name)
        if self.__config.get('password', False):
            os.environ['PGPASSWORD'] = self.__config.get('password')
        createdb_cmd = 'createdb \"{dbname}\" -T template0 -E utf8 --lc-collate C'
        createdb_cmd = self.__format_cmd(createdb_cmd, cnf)
        createdb_cmd = createdb_cmd.format(**cnf)
        shell = spur.LocalShell()
        try:
            shell.run(shlex.split(createdb_cmd))
        except spur.results.RunProcessError as error:
            str_error = utils.decode(error.stderr_output)
            logger.error('Could not create database, error message: %s', str_error)
            raise OperationalError(str_error)
        return database_name

    def restore(self, database_name, dump_name, jobs=False):
        """Restores a database dump in sql plain format, tries to create database if it doesn't
        exist.

        :param database_name: Database where the dump will be restored.
        :type database_name: str
        :param dump_name: Full path and name of the sql dump to restore.
        :type dump_name: str
        :param jobs: How many jobs will be used to restore the database, none by default
        :type dump_name: int
        :returns: If succeeded (True) or not (None).
        :rtype: bool or None
        """
        cnf = self.__config
        cnf.update({'dbname': database_name})

        if self.__config.get('password', False):
            os.environ['PGPASSWORD'] = self.__config.get('password')
        self.__config.update({'dbname': database_name})
        restore_cmd = 'psql {dbname} -w -f {0}'
        if jobs:
            restore_cmd = 'pg_restore -w -j {num} -F d {0} -d {dbname} -x -O -n public'
            cnf.update({'num': str(jobs)})
        restore_cmd = self.__format_cmd(restore_cmd, cnf)
        restore_cmd = restore_cmd.format(dump_name, **cnf)
        shell = spur.LocalShell()
        try:
            shell.run(shlex.split(restore_cmd))
        except spur.results.RunProcessError as error:
            str_error = utils.decode(error.stderr_output)
            logger.error('Could not restore database, error message: %s', str_error)
            self.drop(database_name)
            raise OperationalError(str_error)
        else:
            res = True
        return res

    def create_extension_unaccent(self, database_name):
        """ Creates the unaccent extension in the specified database.
        Keeping this for not braking anything unexpected

        :param database_name: Name of the database where the extension will be created
        """
        logger.info('Creating extension unaccent in the database %s', database_name)
        self.create_extension(database_name, 'unaccent')

    def create_extension(self, database_name, extension_name):
        """ Creates the given extension_name in the database_name.
        The credentials must have permission to create extensions, generally a superuser.

        :param database_name: The database where you want to create the extension
        :param extension_name: The extension name
        :return: None
        """
        config = self.__config.copy()
        config.update({'dbname': database_name})
        logger.info('Creating extension %s in the database %s', extension_name, database_name)
        try:
            with PostgresConnector(config) as db:
                db.execute("create extension {};".format(extension_name))
        except (psycopg2.ProgrammingError, psycopg2.OperationalError) as error:
            error_msg = utils.get_error_message(error)
            allowed_errors = ['permission denied to create extension',
                              'password authentication failed']
            if 'already exists' in error_msg:
                logger.info(error_msg.strip())
                return
            if any(err in error_msg for err in allowed_errors):
                logger.error(
                    'Unable to create the %s extension, authentication for the'
                    ' postgres user in the deployv.conf file failed or the'
                    ' user is not superuser', extension_name
                )
                return
            raise



# coding: utf-8
import logging
import pika
import simplejson as json
from deployv.messaging import basemsg
from deployv.messaging.rabbit.rabbitv import BaseRabbitConfiguration

_logger = logging.getLogger(__name__)


class RabbitSenderV:
    """ A simple rabbit mq interface just pass the config parameters
    and send the message

    """
    def __init__(self, sender_configuration, sender_id):
        assert isinstance(sender_configuration, BaseRabbitConfiguration)
        self.__config = sender_configuration.get_result_object()
        self.__id = sender_id
        self.__connection = None
        self.__channel = None

    def connect(self):
        self.__connection = pika.BlockingConnection(self.__config.parameters)
        self.open_channel()

    def open_channel(self):
        self.__channel = self.__connection.channel()
        self.__channel.queue_declare(queue=self.__config.queue_name, durable=True)
        self.__channel.exchange_declare(exchange=self.__config.exchange_name,
                                        exchange_type='topic')

    def reconnect(self):
        """ Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        # This is the old connection IOLoop instance, stop its ioloop
        if self.__connection:
            self.__connection.ioloop.stop()

        # if not self._closing:

            # Create a new connection
        self.connect()

    def amq_send_message(self, message, send_to=None):
        """ Send a message to the queue using the specified route

        :param message: Message to be delivered
        :param send_to: message route if none is provided will use default
                        passed to the class constructor
        :return: Pika basic_publish return value
        """
        self.connect()
        if self.__config.commit_config:
            routing_key = '%s.commit' % (self.__config.route)
        elif send_to is not None and not self.__config.result_config:
            routing_key = '{route}.{worker}'.format(route=self.__config.route, worker=send_to)
        else:
            routing_key = self.__config.route
        _logger.debug('Sending message (%s): %s', routing_key, message)
        retry = 0
        while retry <= 3:
            try:
                if self.__connection.is_open:
                    self.__channel.basic_publish(self.__config.exchange_name,
                                                 routing_key=routing_key,
                                                 body=message,
                                                 properties=self.__config.properties,
                                                 )
                    retry = 4
                else:
                    self.__connection.add_timeout(self.__config.parameters.socket_timeout,
                                                  self.reconnect)
            except pika.exceptions.ConnectionClosed:
                retry += 1
                _logger.info('Connection failed, retry %s', retry)
                if retry == 3:
                    raise
                self.reconnect()
                # self.__connection.add_timeout(self.__config.parameters.socket_timeout,
                #                               self.reconnect)
        self.__connection.close()

    def send_message(self, message, send_to=None):
        """ Sends a message via rabbit to the specified topic but first checks if the message
        is a MessageV instance

        :param message: MessageV instance
        :param send_to: Node or topic where the message will be sent
        :return: Response from the sender and the original message as string
        """
        assert isinstance(message, (basemsg.BasicMessage, dict))

        if isinstance(message, basemsg.BasicMessage):
            message_str = message.get_message_str()
        else:
            message_str = json.dumps(message,
                                     ensure_ascii=True,
                                     check_circular=True,
                                     encoding='utf-8')
        self.amq_send_message(message_str, send_to)
        return message_str

    def __del__(self):
        if self.__connection and self.__connection.is_open:
            self.__connection.close()



# coding: utf-8

from deployv.messaging.basemsg import BaseWorker
from time import sleep
from random import randint
import logging


_logger = logging.getLogger(__name__)


class RabbitWorker(BaseWorker):
    def __init__(self, configuration_object, sender_class, receiver_class,
                 worker_id, worker_manager_class):
        super(RabbitWorker, self).__init__(configuration_object, sender_class, receiver_class,
                                           worker_id)
        self.worker_manager = worker_manager_class

    def run(self):
        """ Worker method, open a channel through a pika connection and
            start consuming
        """
        _logger.debug("Worker %s - waiting for something to do", self._wid)
        self._receiver.run(self.callback)

    def signal_exit(self):
        """ Exit when finished with current loop """
        self._receiver.stop()

    def kill(self):
        """ This kill immediately the process, should not be used """
        self._receiver.stop()
        self.terminate()
        self.join()

    def should_reconnect(self):
        """Method called after a consumer stops, checks if the consumer was
        stopped manually or if it should reconnect.
        """
        if not self._receiver.should_reconnect:
            return False
        self._receiver.stop()
        wait = randint(3, 7)
        _logger.debug('Waiting %s seconds before retry', wait)
        sleep(wait)
        return True

    def callback(self, channel, method, properties, body):
        """ This method is executed as soon as a message arrives, according to the
            `pika documentation
            <http://pika.readthedocs.org/en/latest/examples/blocking_consume.html>`_. The
            parameters are fixed according to it, even if the are unused
        """
        # always ack the messages that don't pass the filters
        res = True
        message = self.check_message(body)
        if message:
            res = self.worker_manager.start_job(self.execute_rpc, message)
        if res:
            channel.basic_ack(method.delivery_tag)



# coding: utf-8

from deployv.messaging.http.worker import HttpWorker
from deployv.messaging.http.httpv import FileHttpConfiguration, BaseHttpConfiguration
from deployv.messaging.http.receiverv import HttpReceiverV
from deployv.messaging.http.senderv import HttpSenderV


CONFIG_CLASSES = {
    'base': BaseHttpConfiguration,
    'file': FileHttpConfiguration
}


def factory(config_object, worker_id):
    http_worker = HttpWorker(config_object, HttpSenderV, HttpReceiverV, worker_id)
    return http_worker



# -*- coding: utf-8 -*-

"""This module provides different methods to easily verify messages to be sure
they are valid json strings and they comply with the json schema.
"""

import os
import logging
import simplejson
import jsonschema
from six import string_types, binary_type


_logger = logging.getLogger(__name__)


def _get_schema_path(name='base.json'):
    """Builds and returns the json schema file used to validate messages.

    Returns false if the file doesn't exist.

    :param name: The name of the json schema, by default is "base.json".
    :type name: str

    :returns: The json schema file name.
    :rtype: str
    """
    schema_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'json_schemas', name)
    if os.path.isfile(schema_file):
        return schema_file
    return False


def build_schema(model=None):
    """Loads the base schema and it extends it with the specified model schema.

    If there is no specified model, or it doesn't have an associated schema,
    it will return the base schema.

    :param model: The model schema name, for example 'deploy.deploy' or 'deploy_deploy'.
    :type model: str or None

    :returns: The model json schema.
    :rtype: dict
    """
    schema = load_json(_get_schema_path())
    model_name = '{model}.json'.format(model=(model or '').replace('.', '_'))
    model_schema_path = _get_schema_path(model_name)
    if model_schema_path:
        model_schema = load_json(model_schema_path)
        schema.get('properties').update(model_schema.get('properties') or {})
        schema.setdefault('definitions', {}).update(model_schema.get('definitions') or {})
    return schema


def load_json(json, logger=True):
    """Loads a valid json string or file into a dictionary, returning the dict
    object.

    :param json: json string to convert.
    :type json: string or dict
    :param logger: when is False, it does not print the error logger. Default True
    :type logger: bool

    :returns: Returns the loaded dict or False if an error ocurred.
    :rtype: dict or bool

    :Example:

    .. code-block:: python

        json_dict = load_json(json_str)
        if isinstance(json_dict, dict):
            _logger.info("Everything OK")
        else:
            _logger.error("Something wrong")

    :Note: Don't check only using ``if json_dict: ...`` because ``"{}"`` is a
        valid json string and will return a valid object, but an empty object
        is considered `falsy` and the check will fail (unless that is what you
        want).
    """
    if isinstance(json, binary_type):
        json = json.decode()
    if isinstance(json, string_types):
        if os.path.isfile(json):
            return load_json_file(json, logger)
        return load_json_string(json, logger)
    if isinstance(json, dict):
        return json
    if logger:
        error_message = "Invalid type ({type})".format(type=type(json))
        _logger.error("Error loading json: %s", error_message)
    return False


def load_json_string(json, logger=True):
    """Loads a valid json string into a dictionary, returning the dict object.

    :param json: json string to convert.
    :type json: string
    :param logger: when is False, it does not print the error logger. Default True
    :type logger: bool

    :returns: Returns the loaded dict or False if an error ocurred.
    :rtype: dict or bool

    :Example:

    .. code-block:: python

        json_dict = load_json_string(json_str)
        if isinstance(json_dict, dict):
            _logger.info("Everything OK")
        else:
            _logger.error("Something wrong")

    :Note: Don't check only using ``if json_dict: ...`` because ``"{}"`` is a
        valid json string and will return a valid object, but an empty object
        is considered `falsy` and the check will fail (unless that is what you
        want).
    """
    _logger.debug(json)
    try:
        return simplejson.loads(json)
    except ValueError as error:
        if logger:
            _logger.error("Error loading json string: %s", str(error))
        return False


def load_json_file(json, logger=True):
    """Loads a valid json file into a dictionary, returning the dict object.

    :param json: json file to load.
    :type json: string
    :param logger: when is False, it does not print the error logger. Default True
    :type logger: bool

    :returns: Returns the loaded dict or False if an error ocurred.
    :rtype: dict or bool

    :Example:

    .. code-block:: python

        json_dict = load_json_file(json_file_name)
        if isinstance(json_dict, dict):
            _logger.info("Everything OK")
        else:
            _logger.error("Something wrong")

    :Note: Don't check only using ``if json_dict: ...`` because ``"{}"`` is a
        valid json string and will return a valid object, but an empty object
        is considered `falsy` and the check will fail (unless that is what you
        want).
    """
    try:
        with open(json) as json_file:
            return load_json_string(json_file.read(), logger)
    except IOError as error:
        if logger:
            _logger.error("Error loading json file %s: %s", json, str(error))
        return False


def validate_schema(json, schema=None):
    """Validates a json against the valid schema, returning if the validation
    went OK and an error message if any errors occurred.

    Documentation for json schema in https://pypi.python.org/pypi/jsonschema

    :param json: json to validate
    :param schema: the schema to apply to the json. If None is passed it will
        get the default schema defined in this module.
    :type json: str or dict
    :type schema: str or dict or None

    :returns: A dict with "result" (a boolean value if the validation was OK) and
        "error" (a string with the error message if an error occurred).
    :rtype: dict

    :Example:

    .. code-block:: python

        validation = json_helper.validate_schema(json)
        if validation.get("error"):
            _logger.error("Something wrong")
        else:
            _logger.info("Everything OK")
    """
    json = load_json(json)
    schema = build_schema() if schema is None else load_json(schema)
    if not isinstance(json, dict) or not isinstance(schema, dict):
        bad_json = "schema" if not isinstance(schema, dict) else "target object"
        return {"error": "Invalid json in {bad}".format(bad=bad_json)}
    try:
        jsonschema.validate(json, schema)
    except jsonschema.ValidationError as error:
        msg_path = " > ".join(str(item) for item in list(error.absolute_path))
        schema_path = " > ".join(str(item) for item in list(error.absolute_schema_path))
        msg = "Error validating the json: {error} (in: {path}) (schema: {schema})".format(
                   error=str(error), path=msg_path, schema=schema_path)
        _logger.error(msg)
        return {"error": msg}
    return {"result": True}


def save_json(info, filename, ensure_ascii=False):
    """Save info into Json file.

    :param info: Object to be saved
    :param filename: Name of Json file

    :returns: Absolute path of Json file
    :rtype: str
    """
    try:
        with open(filename, 'w') as fout:
            _logger.debug("Opening file %s", filename)
            simplejson.dump(info, fout, sort_keys=True, indent=4,
                            ensure_ascii=ensure_ascii, separators=(',', ':'))
            if not os.path.isabs(filename):
                filename = os.path.abspath(filename)
            _logger.debug("File saved")
    except IOError as error:
        _logger.error(error)
        return False
    return filename



from deployv.helpers import utils
from deployv.base.errors import BuildError, NoSuchImage
from deployv.helpers.clone_oca_dependencies import run as clone_oca_deps
from deployv.instance import SAAS_VERSIONS
from spur.results import RunProcessError
from logging import getLogger
from os import path
from branchesv.branches import load as load_branches
import simplejson as json

_logger = getLogger(__name__)


class CloneV:

    def __init__(self, config, instance):
        self.config = config
        self.instance_manager = instance

    def clone_repositories(self):
        """Create the folder where the repositories will clone,
        then it the ssh key to clone the private repositories and for last
        it clone the repositories and get the variables from the main repository.
        """
        try:
            res_paths = self._create_working_folder()
            self.config.key_ssh_path = utils.deploy_key(
                self.config.instance_config.get('ssh_key'), self.config.temp_working_folder)
            self._clone_dependencies(res_paths)
        except (RunProcessError, BuildError, IOError, NoSuchImage) as error:
            msg_error = utils.get_error_message(error)
            _logger.exception('Could not clone the repos: %s', msg_error)
            return {"error": msg_error}
        return {'result': 'Repositories cloned'}

    def _create_working_folder(self):
        """Creates the working directory where store the repositories cloned.
        """
        instance_path = path.join(self.config.temp_working_folder, 'instance')
        addons_path = path.join(instance_path, 'extra_addons')
        utils.makedir(addons_path)
        return {'instance_path': instance_path, 'addons_path': addons_path}

    def _clone_dependencies(self, paths):
        """Clones the repositories specified in the config and
        search the variables.sh file to set the data in `variables`.

        :param paths: Content the instance folder path where the odoo repository is cloned,
        and the `extra_addons` folder where content all the other repositories cloned.
        :return: True if no error is raised.
        """
        repos = self.config.instance_config.get('repositories')
        res = load_branches(repos, paths['instance_path'], self.config.temp_working_folder,
                            self.config.key_ssh_path)
        cloned = repos
        if res:
            raise BuildError(res.get('msg'))
        varfiles = path.join(paths['instance_path'],
                             self.config.get_mainrepo_path(),
                             'variables.sh')
        variables = utils.read_lines(varfiles)
        self.config.variables.update(utils.parse_variables_to_json(variables))
        if not self.instance_manager.config_variables.get('version'):
            raise BuildError('You need to specify a version in '
                             ' your "variables.sh" file in the repository')
        odoo_url = 'https://github.com/%s.git' % (
            self.instance_manager.config_variables.get('odoo_repo'))
        # TODO: it's easier to create the dict directly with these values instead of
        # adding and calling a method to do so (utils.add_repo())
        extra_repos = [('odoo', odoo_url, 'odoo',
                        self.instance_manager.config_variables.get('odoo_branch'),
                        self.instance_manager.config_variables.get('odoo_commit'))]
        version = self.instance_manager.config_variables.get('version')
        if self.config.instance_config.get('instance_type') != 'production':
            branch = SAAS_VERSIONS.get(version) or version
            extra_repos.append((
                'isolated_addons', 'https://git.vauxoo.com/vauxoo/isolated_addons.git',
                'extra_addons/isolated_addons', branch, False))
        repo_names = [rep.get('name') for rep in repos]
        extra_clone = []
        for name, url, r_path, r_branch, commit in extra_repos:
            if name in repo_names:
                _logger.debug('%s repository is already in the json, skipping', name)
                continue
            extra_clone.append(utils.add_repo(r_branch, name, r_path, url, commit=commit))
        # Clones oca dependencies.
        res_oca = clone_oca_deps(paths['addons_path'], paths['addons_path'], version,
                                 self.config.temp_working_folder, self.config.key_ssh_path)
        if not res_oca[0]:
            raise BuildError(res_oca[1])
        res = load_branches(repos + extra_clone, paths['instance_path'],
                            self.config.temp_working_folder, self.config.key_ssh_path)
        with open(path.join(self.config.temp_working_folder, 'repos_cloned.json'), 'w') as wfile:
            json.dump(cloned + extra_clone, wfile, indent=4, ensure_ascii=False,
                      separators=(',', ':'))
        if res:
            raise BuildError(res.get('msg'))
        return True



# coding: utf-8

import os
import sys
import logging
from setuptools.command.test import test as TestCommand

logging.basicConfig()
_logger = logging.getLogger(__name__)  # pylint: disable=C0103

if os.environ.get('USER', '') == 'vagrant':
    del os.link

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = open('requirements.txt').readlines()
test_requirements = open('test-requirements.txt').readlines()

setup(
    name='deployv',
    version='0.9.191',
    description='Base for all clases and scripts in VauxooTools',
    long_description=readme,
    author='Tulio Ruiz',
    author_email='tulio@vauxoo.com',
    url='https://git.vauxoo.com/deployv/deployv',
    download_url='https://git.vauxoo.com/deployv/deployv',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='vauxootools deployv',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    tests_require=test_requirements,
    cmdclass={
        'test': PyTest
    },
    py_modules=['deployv'],
    entry_points='''
        [console_scripts]
        deployvcmd=deployv.commands.deployvcmd:cli
        workerv=deployv.commands.workerv:run
    ''',
    scripts=[]
)



import tarfile
import zipfile


class DecompressHelper:
    valid_extensions = ('tar.gz', 'tar.bz2', 'zip', 'tar')

    def __init__(self, filename):
        self.fobject = self.get_decompress_object(filename)

    @property
    def support_methods(self):
        suport_method = {
            "tar.gz": lambda fname: tarfile.open(fname, mode='r:gz'),
            "tar.bz2": lambda fname: tarfile.open(fname, mode='r:bz2'),
            "tar": lambda fname: tarfile.open(fname, mode='r:'),
            "zip": lambda fname: zipfile.ZipFile(fname, mode='r')
        }
        return suport_method

    def extractall(self, dest_folder):
        self.fobject.extractall(dest_folder)

    def name_list(self):
        values = []
        if isinstance(self.fobject, tarfile.TarFile):
            values = [i.name for i in self.fobject.getmembers()]
        if isinstance(self.fobject, zipfile.ZipFile):
            values = self.fobject.namelist()
        return values

    def get_decompress_object(self, filename):
        for ext in self.valid_extensions:
            if filename.endswith(ext):
                return self.support_methods[ext](filename)
        return False



# coding: utf-8
import logging
import odoorpc
from docker.errors import NullResource, NotFound
from deployv.base import errors
from deployv.base.extensions_core import EventListenerBase
from deployv.base import postgresv
from deployv.helpers import utils
from deployv.instance import instancev

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class TestPsqlConnection(EventListenerBase):

    class Meta:
        """ This EventListener is made to test the connection with postgres before the instance is
        constructed so it is possible to validate the configuration parameters
        """
        name = 'CheckPsqlConnection'
        event = 'before.create.event'

    def execute(self, obj):
        if obj.instance_manager.config.get('full_stack'):
            logger.info("Instance with internal postgres, skipping postgres check")
            return True
        psql_dict = utils.odoo2postgres(obj.instance_manager.db_config)
        logger.debug('Dict: %s', str(psql_dict))
        psql_dict.update({'dbname': 'postgres'})
        try:
            with postgresv.PostgresConnector(psql_dict) as db_test:
                res = db_test.check_config()
                if res:
                    logger.info('PostgreSQL connection test passed')
                else:
                    logger.info('PostgreSQL connection test failed')
        except (KeyboardInterrupt, errors.GracefulExit):
            raise
        except Exception:  # pylint: disable=W0703
            res = False
            logger.error("Could not connect to the database")
        return res


class TestDockerUser(EventListenerBase):

    class Meta:
        """ This EventListener to test if the user is configured to use docker, pull images, etc
        """
        name = 'CheckDockerUser'
        event = 'before.create.event'

    def execute(self, obj):
        # Check if user can use docker
        res = False
        import getpass    # pylint: disable=C0415
        from docker import APIClient as Client    # pylint: disable=C0415
        cli = Client()
        try:
            cli.containers(all=True)
        except (KeyboardInterrupt, errors.GracefulExit):
            raise
        except Exception as error:  # pylint: disable=W0703
            logger.error(('Current user "%s" cannot connect to docker or'
                          ' docker service is down: %s'),
                         getpass.getuser(), str(error))
        else:
            logger.info('Docker connection test passed for user %s',
                        getpass.getuser())
            res = True
        return res


class InstallTestRepo(EventListenerBase):

    class Meta:
        """ This EventListener is made to install the web_environment_ribbon if it is a test or
        dev instance
        """
        name = 'InstallTestRibbon'
        event = 'after.restore.event'

    def execute(self, obj, returned_value):
        instance_type = obj.instance_manager.instance_type
        logger.debug('Instance type: %s', instance_type)
        if instance_type in instancev.INSTALL_RIBBON\
                and returned_value.get('result', False) \
                and not obj.instance_manager._skip_install_test_ribbon:
            res = True
            if not returned_value.get('result').get('database_name'):
                logger.error('Could not install the web_environment_ribbon, Database not found')
                return False
            if returned_value.get('result').get('critical'):
                logger.error(returned_value.get('result').get('critical'))
                return False
            logger.debug('Installing the web_environment_ribbon')
            install = obj.instance_manager.install_module(
                'web_environment_ribbon_isolated',
                returned_value.get('result').get('database_name')
            )

            if len(install.get('warnings')) >= 1:
                for warn in install.get('warnings'):
                    if 'invalid module names, ignored: web_environment_ribbon' in warn:
                        logger.warning(
                            'Could not install web_environment_ribbon, module not found'
                        )
                        res = False
                        break
        else:
            res = False
        return res


class TestContainer(EventListenerBase):

    class Meta:
        """ This EventListener is made to test if the container is running and the instance is up,
        at this point we cannot assume that there is a database because is just a post event for
        :meth:`deployv.base.commandv.CommandV.create` method and such method does not create
        database
        """
        name = 'CheckContainer'
        event = 'after.create.event'

    def execute(self, obj, returned_value):
        if obj.instance_manager.docker_id is None:
            return False
        cid = obj.instance_manager.docker_id
        inspected = self.inspect(obj)
        if not inspected:
            return False
        if not inspected.get('State').get('Running'):
            logger.error('The container %s is not runnig', cid)
            return False
        # Check if odoo instance is running
        res = obj.instance_manager.exec_cmd('supervisorctl status ' +
                                            obj.instance_manager.supervisor_app)
        if u'RUNNING' not in res:
            logger.error('Odoo instance is not running in container %s', cid)
        else:
            logger.info(
                'Supervisord in container %s reported the instance is running', cid)
        port_binds = obj.instance_manager.basic_info.get('ports').get('8069')
        if not isinstance(port_binds, list):
            host = obj.instance_manager.config.get('domain')
            res = self.test_connection(host, port_binds)
            return res
        for bind in port_binds:
            parts = bind.split(':')
            host = parts[0]
            port = parts[1]
            res = self.test_connection(host, int(port))
            if res:
                return res

    def inspect(self, obj):
        try:
            inspected = obj.instance_manager.inspect()
        except NotFound:
            logger.exception('The provided id/name does not exists')
            return
        except NullResource as error:
            logger.exception('Container is not running: %s',
                             utils.get_error_message(error))
            return
        except errors.NoSuchContainer as error:
            logger.exception('Container does not exists: %s', utils.get_error_message(error))
            return
        return inspected

    def test_connection(self, host, port):
        try:
            odoo = odoorpc.ODOO(host, port=port)
            odoo.db.list()
        except (KeyboardInterrupt, errors.GracefulExit):
            raise
        except Exception as error:  # pylint: disable=W0703
            logger.error('Could not connect to the instance using %s:%s', host, port)
            logger.error(str(error))
            res = False
        else:
            logger.info('Connected to the instance via %s:%s', host, port)
            res = True
        return res

# TODO: This check is not working because it checks that the repos in the JSON
# are the ones that were cloned but doesn't take into account the extra addons
# and the oca dependencies, this needs to be refactored in the task#33028.
# For now, we will disable this check in order to refactor how the update
# is done without problems.

# class TestBranches(EventListenerBase):
#     class Meta:
#         """ This EventListener makes sure that the branches and commits cloned into the new
#         instance are the ones specified in branchesv
#         """
#         name = 'CheckBranches'
#         event = 'after.create.event'
#     def execute(self, obj, returned_value):
#         if returned_value.get('error'):
#             return False
#         try:
#             pre_path = path.join(obj.instance_manager.temp_folder, 'branches.json')
#         except APIError as error:
#             logger.exception('Could not connect to the instance: %s', error.explanation)
#             return False
#         except errors.NotRunning as error:
#             logger.exception('Container is not running: %s',
#                              utils.get_error_message(error))
#             return False
#         except errors.NoSuchContainer as error:
#             logger.exception('Container does not exists: %s', utils.get_error_message(error))
#             return False
#         try:
#             pre_branches = json_helper.load_json(pre_path)
#         except IOError as error:
#             logger.warning(utils.get_error_message(error))
#             logger.warning('Maybe no branches was provided in the config (this is not an error)')
#             return True
#         home = obj.instance_manager.config.get('env_vars').get('odoo_home')
#         instance_dir = path.join(home, 'instance/')
#         post_path = path.join('/tmp', 'post_process.json')
#         try:
#             obj.instance_manager.exec_cmd('branchesv save -p {i} -f {p}'.format(i=instance_dir,
#                                                                                 p=post_path))
#         except errors.NotRunning as error:
#             logger.exception('Could not load the branches: %s',
#                              utils.get_error_message(error))
#             return False
#         post_branches = json_helper.load_json(
#             path.join(
#                 obj.instance_manager.temp_folder,
#                 'post_process.json'
#             ))
#         for value in pre_branches:
#             if not value['commit']:
#                 url_dict = value['repo_url']
#                 url = url_dict.get('origin')
#                 cmd = ('git ls-remote {remote} HEAD'.format(remote=url))
#                 result = obj.instance_manager.exec_cmd(cmd)
#                 latest = result.split()
#                 value['commit'] = latest[0].strip()
#             if value not in post_branches:
#                 logger.info('Repository not cloned: %s', value['name'])
#             else:
#                 logger.info('Repository %s successfully cloned', value['name'])





class DeployVException(Exception):
    msg = None

    def __init__(self, message, *args, **kwargs):
        self.msg = message
        super(DeployVException, self).__init__(message, *args, **kwargs)

    def __str__(self):
        return self.msg


class NotRunning(DeployVException):
    def __init__(self, container):
        super(NotRunning, self).__init__('Container {} is not running'.format(container))


class NoSuchContainer(DeployVException):
    def __init__(self, container):
        super(NoSuchContainer, self).__init__('No such container or id: {}'.format(container))


class FileNotFoundInContainer(DeployVException):
    def __init__(self, container, filename):
        super(FileNotFoundInContainer, self).__init__('File {0} not found in the {1} container'
                                                      .format(filename, container))


class ErrorPort(DeployVException):
    def __init__(self, message):
        super(ErrorPort, self).__init__(message)


class BuildError(DeployVException):
    def __init__(self, message):
        super(BuildError, self).__init__(message)


class NoSuchBranch(DeployVException):
    def __init__(self, branch, repo):
        super(NoSuchBranch, self).__init__(('Could not find the specified remote branch'
                                            ' {branch} in repo: {repo}').format(branch=branch,
                                                                                repo=repo))


class MethodNotImplemented(DeployVException):
    def __init__(self, message):
        super(MethodNotImplemented, self).__init__(message)


class NoSuchImage(DeployVException):
    def __init__(self, image_name):
        super(NoSuchImage, self).__init__('Image not found: {image}'
                                          .format(image=image_name))


class DumpError(DeployVException):
    def __init__(self, message):
        super(DumpError, self).__init__(message)


class GracefulExit(DeployVException):
    def __init__(self, message):
        super(GracefulExit, self).__init__(message)


class CommandError(DeployVException):
    def __init__(self, cmd, error):
        super(CommandError, self).__init__(('Failed to execute the command {cmd}: {error}'
                                            .format(cmd=cmd, error=error)))


class SupervisorStatusError(DeployVException):
    def __init__(self, message):
        super(SupervisorStatusError, self).__init__(message)


class ClientAPIError(DeployVException):
    def __init__(self, message):
        super(ClientAPIError, self).__init__(message)


class ParameterError(DeployVException):
    def __init__(self, message):
        super(ParameterError, self).__init__(message)


class NoItemFound(DeployVException):
    def __init__(self, message):
        super(NoItemFound, self).__init__(message)



from multiprocessing import Process, Manager, active_children
import multiprocessing.managers as mpmanagers
from queue import Full, Empty
import logging


_logger = logging.getLogger(__name__)

# multiprocessing has a bug in some versions of python where the AutoProxy
# method is called with an extra parameter (manager_owned), we are wrapping
# that method so we can receive (and ignore) that extra parameter without
# raising an exception. More info in `https://bugs.python.org/issue30256`
ORIGINAL_AUTOPROXY = mpmanagers.AutoProxy


def fixed_autoproxy(token, serializer, manager=None, authkey=None,
                    exposed=None, incref=True, manager_owned=False):
    return ORIGINAL_AUTOPROXY(token, serializer, manager, authkey,
                              exposed, incref)


mpmanagers.AutoProxy = fixed_autoproxy


class WorkerManager:

    def __init__(self, worker_type, workers):
        # If the workers will only be used to report the status to orchest,
        # just start one worker.
        self._workers = workers if worker_type != 'Status worker' else 1
        self.available_workers = Manager().Queue(workers)
        _logger.info('Initializing the workers for the %s', worker_type)
        self.start_workers()

    def start_workers(self):
        """ Starts WorkerUnit instances ready to attend the messages equal
        to the amount of desired workers.
        """
        for _ in range(self._workers):
            self.add_worker()

    def add_worker(self):
        """ Adds a new worker to the queue of workers, used to create the
        initial "pool"" of workers, and to replace the workers that are removed
        when a new message arrives.
        """
        try:
            self.available_workers.put(WorkerUnit, block=False)
        except Full:
            _logger.info("Can't add more workers, limit reached: %s",
                         self._workers)
        _logger.info('New worker started')

    def start_job(self, func, data):
        """ Takes a new worker from the "pool" of workers and
        starts a new process in  the background where the worker attends a job.

        :param func: function to be executed by the worker.
        :param data: Dictionary with the data of the message received to be passed to
            the function.

        :return: True it the job started successfully, False otherwise.
        """
        # Remove the defunct processes of the workers that already finished
        # before starting a new process
        active_children()
        try:
            worker = self.available_workers.get(block=False)
            worker_inst = worker(manager=self)
        except Empty:
            _logger.debug('No available workers for this job')
            return False
        _logger.info('Working on the message:\n%s', data)
        worker_process = Process(target=worker_inst.work, args=(func, data,))
        worker_process.start()
        return True


class WorkerUnit:

    def __init__(self, manager):
        self._manager = manager

    def work(self, func, data):
        """ Executes the provided function with the provided data and adds
        a new worker to the pool after it finishes.
        """
        func(data)
        self._manager.add_worker()
        _logger.info('Job done')



# coding: utf-8

from os import path
import base64
from datetime import datetime
import uuid
from deployv.helpers import json_helper, utils, configuration_helper
from deployv.base.errors import GracefulExit
from deployv.instance.instancev import DummyInstanceV
import simplejson as json
import time
import logging
try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser


_logger = logging.getLogger('deployv')
TRIGGER_MAPPINGS = {
    'reload_config': 'reload_config_trigger',
    'build_image': 'build_image_trigger',
    'create': 'create_trigger',
    'commit_history': 'commit_history_trigger',
    'update_instance': 'create_trigger',
    'restart_instance': 'instance_restart_trigger',
    'restart_container': 'container_restart_trigger',
    'change_passwords': 'change_passwords_trigger',
    'delete_instance': 'destroy_trigger',
    'backupdb': 'backup_trigger',
    'deactivate_backup': 'deactivate_backup_trigger',
    'push_image': 'push_images_trigger',
    'update_production': 'update_prod_trigger',
    'compare_databases': 'compare_databases',
}


class BaseWorker:

    def __init__(self, configuration_object, sender_class, receiver_class, worker_id):
        self.config_obj = configuration_object.config
        self._wid = worker_id
        self._wtype = configuration_object.wtype
        self.count = 0
        self._config_class = self.config_obj
        self._receiver = receiver_class(configuration_object,
                                        self.config_obj.deployer.get('node_id'))
        self._sender = sender_class(configuration_object, self.config_obj.deployer.get('node_id'))

    def run(self):
        """ This is the start up method
        """
        raise NotImplementedError

    def exit(self):
        """ Exit worker, blocks until worker is finished and dead """
        self.signal_exit()
        while self.is_alive():  # checking `is_alive()` on zombies kills them
            time.sleep(1)

    def kill(self):
        """ This kill immediately the process, should not be used """
        raise NotImplementedError

    def check_message(self, message):
        """ Check if the message is properly formed and can be parsed, if nt a message with the
            error will be generated

        :param message: The message to be checked
        :return: The message object (:class:`~deployv.messaging.messagev.BasicMessage`)
            if no error, else False
        """
        message = json_helper.load_json(message)
        if not isinstance(message, dict):
            return False
        res = BasicMessage(message)
        model_schema = json_helper.build_schema(message.get('res_model'))
        validation = json_helper.validate_schema(message, schema=model_schema)
        validate_cmd = utils.validate_command(message.get('message_body') or {},
                                              self.config_obj.deployer.get('server_type'))
        validate_error = validation.get('error') or validate_cmd.get("error")
        if validate_error:
            res.set_message_body({
                'error': validate_error,
                'command': (message.get('message_body') or {}).get('command'),
            }, message_type='error')
            result = res.build_message()
            self._sender.send_message(result)
            return False
        return res

    def execute_rpc(self, message):
        _logger.info('Executing rpc for message in worker %s', self._wid)
        _logger.debug('%s worker received "%s"', self._wid,
                      json.dumps(message.original_message, sort_keys=True, indent=4))
        node_id = self.config_obj.deployer.get('node_id')
        if self._wtype != 'commit' and message.receiver_node_id != node_id:
            _logger.error('Message in the wrong queue, does not match my wid: %s', node_id)
            return False
        self._sender.send_message(message.get_ack_message())
        _logger.debug('Ack sent')
        module = message.message_body.get('module')
        command_name = message.message_body.get('command')
        parameters = message.message_body.get('parameters')
        classes = {'commandv': 'CommandV'}
        if not parameters.get('group_config'):
            parameters.update({'group_config': {}})
        parameters.get('instance', {}).get('config', {}).update({
            'db_owner': self.config_obj.postgres.get('db_user'),
            'db_owner_passwd': self.config_obj.postgres.get('db_password')
        })
        parameters.get('container_config', {}).update({
            'working_folder': self.config_obj.deployer.get('working_folder')
        })
        if not parameters.get('container_config', {}).get('domain'):
            parameters.get('container_config', {}).update({
                'domain': self.config_obj.deployer.get('domain')})
        action_name = 'regenerate' if command_name == 'delete_instance' and parameters.get(
            'group_config').get('create_container') else command_name
        default_params = {
            'backup_src': self.config_obj.deployer.get('backup_folder'),
            'temp_folder': self.config_obj.deployer.get('temp_folder'),
            'max_instances': self.config_obj.deployer.get('max_instances'),
            'docker_url': self.config_obj.deployer.get('docker_url'),
            'use_nginx': self.config_obj.deployer.get('use_nginx'),
            'nginx_folder': self.config_obj.deployer.get('nginx_folder'),
            'res_id': message.res_id,
            'sender_node_id': message.receiver_node_id or node_id,
            'deploy_id': message.deploy_id,
            'receiver_node_id': message.sender_node_id,
            'res_model': message.res_model,
            'model': message.model,
            'user_id': message.user_id,
            'action_name': action_name,
            'response_to': message.message_id,
        }
        load_defaults = parameters.get('group_config', {}).get('load_defaults', False)
        for param, value in default_params.items():
            if param not in parameters.get('group_config').keys():
                parameters.get('group_config').update({param: value})
        # Create a new config object for each command instead of using the same for every thread
        command_config = configuration_helper.DeployvConfig(deploy_config=parameters,
                                                            load_defaults=load_defaults)
        command_config.update_configuration(worker_config=self.config_obj.worker_config)
        message_body = {
            'command': command_name,
            'module': module,
        }
        instance_class = DummyInstanceV if self._wtype == 'commit' else False
        if module == 'commandv':
            module_base = __import__('deployv.base', fromlist=[str(module)])
            module_command = getattr(module_base, module)
            command_obj = getattr(module_command, classes.get(module))(
                command_config, instance_class)
            _logger.debug('Parameters: %s',
                          json.dumps(parameters, sort_keys=True, indent=4))
            trigger = TRIGGER_MAPPINGS.get(command_name)
            if not trigger:
                message_body.update({'error': 'Command {cmd} does dot exits in module {module}'
                                    .format(cmd=command_name, module=module)})
                message.set_message_body(message_body, message_type='error')
                response = message.build_message()
                self._sender.send_message(response)
                return
            starting_trigger = getattr(command_obj, trigger)
            try:
                starting_trigger()
            except Exception as error:  # pylint: disable=W0703
                message_body.update(({'error': utils.get_error_message(error)}))
                command_obj.send_message_error(message_body)
                command_obj.clean_working_folders()
                if isinstance(error, (KeyboardInterrupt, GracefulExit)):
                    raise
                _logger.exception(message_body.get('error'))
            _logger.info('%s worker done', self._wid)


class BasicMessage:
    """ Basic unit of communication that will be used for remote communications, you need at least
    specify the sender, receiver and a message body (command, error, status, etc)::

        >>> message = basemsg.BasicMessage()
        >>> message.sender_node_id = 'me01'
        >>> message.receiver_node_id = 'you01'
        >>> res = message.build_message()

    The :func:`~deployv.base.messagev.BasicMessage.build_message` method generates the proper
    message according to the defines schema for the communication (check test_messagev.py
    in the tests folder)

    """

    def __init__(self, message=None):
        envelope = {
            'version': '0.1',
            'message_id': str(uuid.uuid1()),
            'res_id': None,
            'deploy_id': None,
            'orchest_pipe_id': None,
            'sender_node_id': None,
            'receiver_node_id': None,
            'timestamp': None,
            'response_to': None,
            'message_body': dict(),
            'res_model': None,
            'model': None,
            'user_id': None
        }
        self.__dict__ = envelope.copy()
        self._envelope = envelope.copy()
        self._files = list()
        self._message_types = ['error', 'result', 'parameters']
        if message is None:
            self._original_message = None
        elif isinstance(message, str):
            self._original_message = json.loads(message)
        elif isinstance(message, dict):
            self._original_message = message.copy()
        else:
            raise TypeError('Message is in a unsupported type {}'.format(type(message)))
        if self._original_message is not None:
            for key in envelope:
                setattr(self, key, self._original_message.get(key))

    @property
    def original_message(self):
        return self._original_message

    def attach_file(self, file_name, mime_type=None):
        assert path.isfile(file_name)
        self._files.append((file_name, mime_type))

    def set_message_body(self, message_body, message_type):
        """ Define the message body to be send

        :param message_body: Message body according to the documentation
        :param message_type: Error, ack, response or parameters
        :return: None
        """
        if message_type not in self._message_types:
            raise ValueError('Message type must be one of: {}'.format(str(self._message_types)))
        self.message_body = message_body

    def set_command(self, module_command, parameters):
        """ Helper method to create the message for a command

        :param module_command: Execute the command from the specified module.
            Must be in the format module.command
        :param parameters: The parameters that will be passed to the command in a dict
        :return: None
        """
        assert len(module_command.split('.')) == 2
        assert isinstance(parameters, dict)
        res = {
            'module': module_command.split('.')[0],
            'command': module_command.split('.')[1],
            'parameters': parameters
        }
        self.message_body = res

    def build_message(self):
        """ Builds the message with the provided properties in the class and according to the type

        :return: The formatted message in a dict
        """
        res = dict()
        attachments = list()
        for key in self._envelope:
            res.update({key: getattr(self, key)})
        res.update({
            'timestamp': datetime.utcnow().isoformat()})
        for fname in self._files:
            with open(fname(0)) as fname_descriptor:
                coded_file = utils.decode(base64.b64encode(utils.encode(fname_descriptor.read())))
            attachments.append({
                'file_name': path.basename(fname(0)),
                'file': coded_file,
                'type': fname(1)
            })
        if self._original_message:
            res.update({
                'receiver_node_id': self._original_message.get('sender_node_id'),
                'sender_node_id':  self._original_message.get('receiver_node_id'),
                'res_model': (self._original_message.get('res_model') or
                              self._original_message.get('model'))
            })
        if res.get('message_body').get('error', False) and attachments:
            res.get('message_body').get('error').update({'attachments': attachments})
        elif res.get('message_body').get('result', False) and attachments:
            res.get('message_body').get('result').update({'attachments': attachments})
        return res

    def get_ack_message(self, send_to=None):
        """ When a message is provided in the constructor this method generates an ack response
        with the proper parameters

        :param send_to: message route if none is provided will use default
                        passed to the class constructor
        :return: The ack message
        """
        if self._original_message is None:
            res = self.build_message()
        else:
            res = self._original_message.copy()
        res.update({
            'receiver_node_id': res.get('receiver_node_id'),
            'sender_node_id': res.get('sender_node_id') if send_to is None else send_to,
            'res_model': res.get('res_model'),
            'model': res.get('model'),
            'res_id': res.get('res_id'),
            'deploy_id': res.get('deploy_id'),
            'timestamp': self.timestamp,
            'response_to': res.get('message_id'),
            'message_id': str(uuid.uuid1()),
            'message_body': {
                'module': res.get('message_body').get('module'),
                'command': res.get('message_body').get('command'),
                'ack': {
                    'message': 'Message received'
                }
            }
        })
        return BasicMessage(message=res)

    def get_message_str(self):
        """ A simple helper method that returns the message as a str but being sure that it will
        be safe for rabbitmq

        :return: The string representing the message
        """
        return json.dumps(self.build_message(),
                          ensure_ascii=True,
                          check_circular=True,
                          encoding='utf-8')

    def __repr__(self):
        return self.get_message_str()


class BaseFileConfig:
    def __init__(self, config_file, result=False):
        assert path.isfile(config_file)
        self.config = ConfigParser()
        self._result = result



# coding: utf-8

from __future__ import division
import os
import re
import logging
import logging.handlers
import shutil
import shlex
import spur
import random
import datetime
import base64
import tarfile
import paramiko
import csv
from six import string_types, binary_type
import binascii
from string import digits, ascii_letters  # pylint: disable=deprecated-module
from deployv.helpers import container, json_helper
from deployv.helpers.decompress_helper import DecompressHelper
from deployv import base
import deployv_static
from jinja2 import Environment, FileSystemLoader
from errno import EEXIST
from deepdiff import DeepDiff
from stat import S_IREAD, S_IWRITE
import socket
import fcntl
import struct
import ipaddress

_AVAILABLES = {
    "production": ['update_production', 'post_process', 'restart_instance',
                   'restart_container', 'reload_config'],
    "test": [],
    "updates": [],
    "develop": [],
}

IMPLEMENTED_MSG_ERROR = ('The class you are instantiating does '
                         'not have this method implemented')

# pylint: disable=C0103
logger = logging.getLogger(__name__)


def copy_list_dicts(lines):
    res = []
    for line in lines:
        dict_t = {}
        for keys in line.keys():
            dict_t.update({keys: line[keys]})
        res.append(dict_t.copy())
    return res


def setup_deployv_logger(name='deployv', level=logging.DEBUG, log_file=None):
    """Configures a logger, setting the level and a stream handler by default.
    If a log file name is passed, a file handler is also created.

    By default it configures the 'deployv' logger, so the config applies
    automatically to every other logger that belongs to a submodule.
    (For example: 'deployv.commands.deployvcmd')

    Important note: To avoid multiple configurations or overriding other loggers,
    the loggers should be called using 'logging.getLogger(__name__)', so the
    logger is named like the file where is being used. This also facilitates that
    every logger inside the 'deployv' module gets the correct parent settings.

    This function returns the configured logger, but it should be used to
    configure the parent logger ('deployv') and only use a child logger from
    another module.
    """
    dv_logger = logging.getLogger(name)
    dv_logger.propagate = False
    dv_logger.setLevel(level)
    dv_logger.handlers = []
    formatter = logging.Formatter(fmt="%(asctime)s - %(levelname)-5s - "
                                      "%(name)s.%(module)s.%(funcName)s - "
                                      "%(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    dv_logger.addHandler(stream_handler)
    if log_file:
        file_handler = logging.handlers.RotatingFileHandler(log_file)
        file_handler.setFormatter(formatter)
        dv_logger.addHandler(file_handler)
    return dv_logger


def is_iterable(obj):
    """ Method that verifies if an object is iterable and not a string, example:

    .. doctest::

        >>> from deployv.helpers import utils
        >>> utils.is_iterable(1)
        False
        >>> utils.is_iterable([1, 2, 3])
        True

    :param obj: Any object that will be tested if is iterable
    :return: True or False if the object can be iterated
    """
    return hasattr(obj, '__iter__') and not isinstance(obj, string_types + (binary_type,))


def load_default_config():
    """ Loads the default configuration file.

    :return: Dictionary with the default configuration
    """
    default_config_path = deployv_static.get_template_path('default_config.json')
    default_config = json_helper.load_json(default_config_path)
    return default_config


def merge_config(config_file, keys_file=False, branches=False):
    """Merges keys_file and branches into config_file and returns a single json
        if no keys_file or branches is given just loads the config file

    :param config_file: Config file in json format
    :param keys_file: Param with the key file name and full path to it
    :param branches: Branches file to be used
    :return: Dict with complete json configuration after saving it into the config_file file
    """
    json_config = json_helper.load_json(config_file)
    if not isinstance(json_config, dict):
        return
    default_config = load_default_config()
    new_config = merge_dicts(default_config, json_config)
    changed = False
    if branches:
        json_branches = json_helper.load_json(branches)
        new_config.get("instance").update({"repositories": json_branches})
        changed = True
    if keys_file:
        encoded_key = generate_attachment(keys_file)
        new_config.get("instance").update({"ssh_key": encoded_key})
        changed = True
    if changed:
        json_helper.save_json(new_config, config_file)
    return new_config


def list_backups(file_dir, prefix):
    """List all files which start with prefix and sort
        them by date

    :param file_dir: Directory where backups are stored
    :param prefix: Prefix files to seek
    :return: sorted list
    """
    items = []
    for each in os.listdir(file_dir):
        if (prefix in each) and os.path.isfile(os.path.join(file_dir, each)):
            items.append(str(each))
    items.sort(reverse=True)
    return items


def clean_files(files):
    """Remove unnecessary and temporary files.

    :param files: A list or a str of absolute or relative paths thar will be erased
    """
    if not files:
        return
    items = files if is_iterable(files) else [files]
    for item in items:
        fname = item[0] if is_iterable(item) else item
        if fname != "/":
            logger.info('Removing %s', fname)
            if os.path.isfile(fname):
                os.remove(fname)
            elif os.path.isdir(fname):
                shutil.rmtree(fname)
        else:
            logger.error(
                "Invalid target path: '/'. Are you trying to delete your root path?")


def resume_log(log_lines):
    """Gets the log lines from -u (modules or all) and parse them to get the totals
        according to the filters dict

    :param log_lines: each element of the list is a log line
    :return: dict with key filters as keys and a list with all matched lines
    """
    def critical(line):
        criteria = re.compile(r'.*\d\sCRITICAL\s.*')
        return criteria.match(line)

    def errors(line):
        criteria = re.compile(r'.*\d\sERROR\s.*')
        return criteria.match(line)

    def warnings_trans(line):
        criteria = re.compile(
            r'.*\d\sWARNING\s.*no translation for language.*')
        return criteria.match(line)

    def import_errors(line):
        criteria = re.compile(r'^ImportError.*')
        return criteria.match(line)

    def warnings(line):
        criteria = re.compile(r'.*\d\sWARNING\s.*')
        return criteria.match(line) and 'no translation for language' not in line

    filters = {
        'critical': critical,
        'errors': errors,
        'warnings': warnings,
        'warnings_trans': warnings_trans,
        'import_errors': import_errors
    }
    res = {name: [] for name in filters}
    for line in log_lines:
        stripped_line = line.strip()
        for name, criteria in filters.items():
            if criteria(stripped_line):
                res.get(name).append(stripped_line)
                break

    return res


def get_strtime():
    """ Returns time stamp formatted as follows::

        %Y%m%d_%H%M%S

        So all backups and files that require this in the name will have the same format,
        if changes will be in one place, here.

    :return: Formatted timestamp
    """
    return datetime.datetime.now().strftime("%Y%m%d_%H%M%S")


def get_decompress_object(file_name):
    """ Returns an object to extract file_name*, only tar.gz and tar.bz2 are supported now:

    .. doctest::

        >>> from deployv.helpers import utils
        >>> utils.get_decompress_object('filename.tar.bz2')
        (<bound method TarFile.open of <class 'tarfile.TarFile'>>, 'r:bz2')

    Or if you pass a tar.gz file:

    .. doctest::

        >>> from deployv.helpers import utils
        >>> utils.get_decompress_object('filename.tar.gz')
        (<bound method TarFile.open of <class 'tarfile.TarFile'>>, 'r:gz')

    :param file_name: File name to be extracted
    :return: Tuple with the object and mode
    """
    if file_name.endswith('tar.gz'):
        fobject, modestr = tarfile.open, 'r:gz'
    elif file_name.endswith('tar.bz2'):
        fobject, modestr = tarfile.open, 'r:bz2'
    elif file_name.endswith('tar'):
        fobject, modestr = tarfile.open, 'r:'
    else:
        raise RuntimeError('Unknown file format "{}"'.format(file_name))
    return fobject, modestr


def decompress_files(name, dest_folder):
    """ Decompress a file, set of files or a folder compressed in tar.bz2 format

    :param name: Compressed file name (full or relative path)
    :param dest_folder: Folder where the decompressed files will be extracted
    :return: The absolute path to decompressed folder or file
    """
    assert os.path.exists(name)
    logger.debug("Decompressing file: %s", name)
    if os.path.isdir(name):
        return name
    logger.debug('Extracting %s into %s', name, dest_folder)
    fobject = DecompressHelper(name)
    try:
        fobject.extractall(dest_folder)
    except (EOFError, IOError) as error:
        logger.exception('Error uncompressing file %s', get_error_message(error))
        raise
    name_list = fobject.name_list()
    base_folder = dest_folder
    for fname in name_list:
        if (os.path.basename(fname) in
           ['dump.sql', 'database_dump.b64', 'database_dump.sql', 'database_dump']):
            base_folder = os.path.dirname(fname)
            break

    logger.debug("Destination folder: %s", dest_folder)
    logger.debug("Bakcup folder: %s", base_folder)
    if name.endswith(DecompressHelper.valid_extensions):
        fname = os.path.basename(name)
        dest_folder = os.path.join(dest_folder, base_folder)
    logger.debug("Destination folder: %s", dest_folder)
    return dest_folder


def odoo2postgres(odoo_config):
    """ This is a helper to convert from a odoo configuration dict to a postgres configuration dict
    using libpq
    format: http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-PARAMKEYWORDS,
    and also can map env vars gotten from a container

    :param odoo_config: Odoo database configuration
    :return: Dict with psql configuration
    """
    mapping = {
        'user': ['db_user', 'DB_USER'],
        'host': ['db_host', 'DB_HOST'],
        'port': ['db_port', 'DB_PORT'],
        'password': ['db_password', 'DB_PASSWORD'],
        'dbname': ['db_name', 'DB_NAME']
    }
    res = {}
    for psql_key, psql_value in mapping.items():
        for odoo_key, odoo_value in odoo_config.items():
            if odoo_key in psql_value:
                res.update({psql_key: odoo_value})
    return res


def generate_dbname(config, backup_name=False, prefix=False):
    """ Generates a database name for a test/dev instance based on the container name and
    backup used for the restoration. The created name will have the same timestamp that the one
    used, but if no backup name is supplied or the backup does not have a timestamp in the name
    the actual date will be used

    :param config: Instance configuration object
    :param backup_name: Optional backup name to be restored
    :return:
    """
    if not prefix:
        prefix = container.generate_prefix(config)
    if backup_name:
        name = re.search(r"_(\d{8}_\d{6})", backup_name)
        if name:
            name = name.group(1)
        else:
            name = get_strtime()
    else:
        name = get_strtime()
    res = '{prefix}_{date}'.format(prefix=prefix, date=name)
    return res


def generate_backup_name(database_name, reason=False, prefix=False):
    """Generates the backup name according to the following standard:
       database_name_reason_YYYYmmdd_HHMMSS
       If reason is none:
       database_name_YYYYmmdd_HHMMSS
    """
    if reason and not prefix:
        res = '%s_%s_%s' % \
            (database_name, reason, datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))
    elif prefix and not reason:
        res = '%s_%s' % \
            (prefix, datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))
    elif prefix and reason:
        res = '%s_%s_%s' % \
            (prefix, reason, datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))
    else:
        res = '%s_%s' % (
            database_name, datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))
    return res


def decode_b64_file(src, dst):
    """ Read src base64 encoded file and output its content to dst file
    :param src: Source file to read
    :param dst: Destination file
    """
    with open(src, 'r') as source_file:
        with open(dst, 'w') as destination_file:
            for line in source_file:
                destination_file.write(decode(base64.b64decode(line)))


def compress_files(name, files, dest_folder=None, cformat='bz2'):
    """ Compress a file, set of files or a folder in the specified cforma

    :param name: Desired file name w/o extension
    :param files: A list with the absolute o relative path to the files
                      that will be added to the compressed file
    :param dest_folder: The folder where will be stored the compressed file
    :param cformat: Desired format for compression, only supported bz2 and gz
    """
    if not dest_folder:
        dest_folder = '.'
    if cformat not in ['bz2', 'gz', 'tar']:
        raise RuntimeError('Unknown file format "{}"'.format(cformat))
    ext = ''
    if cformat == 'gz':
        fobject, modestr = tarfile.open, 'w:gz'
        ext = 'tar.gz'
    elif cformat == 'tar':
        fobject, modestr = tarfile.open, 'w:'
        ext = 'tar'
    elif cformat == 'bz2':
        fobject, modestr = tarfile.open, 'w:bz2'
        ext = 'tar.bz2'
    logger.debug("Generating compressed file: %s in %s folder",
                 name, dest_folder)

    bkp_name = '{0}.{1}'.format(name, ext)
    full_tmp_name = os.path.join(
        dest_folder,
        '._{}'.format(bkp_name)
    )
    full_name = os.path.join(dest_folder, bkp_name)

    with fobject(full_tmp_name, mode=modestr) as tar_file:
        for fname in files:
            if is_iterable(fname):
                tar_file.add(fname[0], os.path.join(name, fname[1]))
            else:
                basename = os.path.basename(fname)
                tar_file.add(fname, os.path.join(name, basename))
    shutil.move(full_tmp_name, full_name)
    return full_name


def generate_attachment(file_name):
    """ Helper that generates a file with base 64 encoded content
    to be used as an attachment in the messages returned by deployv

    :param file_name: Full path and name of the file to be attached
    :return: File content in b64 format
    """
    with open(file_name) as attch_file:
        res = decode(base64.b64encode(encode(attch_file.read())))
    return res


def get_error_message(exception_obj):
    """Get the message error from exception object or dict.

    :param exception_obj: the exception object or dict exception where get the message.
    :type: object, dict
    :return: A string containing the exception message.
    :rtype: str
    """
    error_attrs = ['stderr_output', 'explanation', 'msg', 'strerror',
                   'message', 'error_message', 'data', 'name']

    if isinstance(exception_obj, str):
        return str(exception_obj)
    for attr in error_attrs:
        if isinstance(exception_obj, dict) and exception_obj.get(attr):
            return exception_obj[attr]
        if not hasattr(exception_obj, attr):
            continue
        msg = getattr(exception_obj, attr)
        if not msg:
            continue
        msg = json_helper.load_json(msg, False) or msg
        if isinstance(msg, dict):
            msg = msg.get("error") or msg.get("message") or msg
        if isinstance(msg, bytes):
            msg = msg.decode()
        return msg
    return repr(exception_obj)


def makedir(path_name):
    path_name = os.path.expanduser(path_name)
    try:
        os.makedirs(path_name)
    except OSError as error:
        if error.errno != EEXIST:
            raise
    return path_name


def clone_repo(repo, branch, path):
    shell = spur.LocalShell()
    try:
        shell.run(
            shlex.split('git clone -b {branch} --single-branch --depth=1 {repo} {path}'.format(
                branch=branch, repo=repo, path=path)))
    except spur.results.RunProcessError as error:
        if 'Could not find remote branch' in get_error_message(error):
            raise base.errors.NoSuchBranch(branch, repo)
        if 'already exists and is not an empty directory' not in get_error_message(error):
            raise


def merge_dicts(original, new):
    """ Updates key by key a dictionary (original) with the values from another dictionary (new)
        if the value of a key in the new dictionay and the value of the same key in the original
        dictionary are dictionaries, it merges both dictionaries first, if they are lists, it
        appends the values of both lists, otherwise it will replace the value of the original
        dictionary with the value in the new one
    """
    res = original.copy()
    for new_key, new_value in new.items():
        original_value = original.get(new_key)
        if isinstance(original_value, dict) and isinstance(new_value, dict):
            new_dict = merge_dicts(original_value, new_value)
            res.update({new_key: new_dict})
        else:
            res.update({new_key: new_value})
    return res


def clean_string(string):
    """ When creating an image from a branch containing any special char like ., # or $
        or containing upper case chars docker shows an error because those are not allowed chars

    :param string: The string you want to clean
    :return: A string without the invalid chars and without upper case chars
    """
    logger.debug('Clean string %s', string)
    res = re.sub(r"[\.#\$]", "", string)
    return res.lower()


def validate_external_file(file_path):
    """ Helper method that makes sure the a file is plain text and not too big
    so it can be added to the config json.

    :param file_path: Path to the external file we want to add to the config.
    :return: True if the file is an acceptable candidate to be added to the config,
             False otherwise.
    """
    file_size = os.path.getsize(file_path)
    if os.path.splitext(file_path)[1] == '.json':
        if not json_helper.load_json(file_path):
            return False
    if file_size > 500000:
        logger.error('%s is not a valid file, it\'s size shouldn\'t be more than 500kb,'
                     ' file size: %s', file_path, file_size)
        return False
    return True


def random_string(length):
    """ Generates a random string consisting of numbers and chars with the specified length
    """
    res = u''.join(random.choice(ascii_letters+digits) for letter in range(length))
    return res


def decode(string, errors='replace'):
    if isinstance(string, binary_type) and not isinstance(string, string_types):
        return string.decode(encoding=base.CHARSET, errors=errors)
    return string


def encode(string, errors='replace'):
    if isinstance(string, string_types) and not isinstance(string, binary_type):
        return string.encode(encoding=base.CHARSET, errors=errors)
    return string


def parse_url(url):
    """ Parses an url and returns the parts that we are interested in:
        port, domain, user, destination path

    :param url: the url to be parsed, the following format is expected:
                protocol://[user@]domain.com[:port]/remote/path
    :return: Dict with the parsed values
    """
    match = re.match(
        r'^(?P<prot>\w+)://((?P<user>\w+)@)?(?P<dom>[\w|.|-]+)(:(?P<port>\d+))?/(?P<path>.*)$',
        url)
    res = dict(
        protocol=match.group('prot'),
        user=match.group('user'),
        domain=match.group('dom'),
        port=match.group('port'),
        folder=match.group('path')
    )
    return res


def upload_scp(filename, credentials, retry=3):
    logger.info('Uploading %s using SFTP', filename)
    port = credentials.get('port') if credentials.get('port') else 22
    private_key = paramiko.RSAKey.from_private_key_file(
        os.path.expanduser(os.path.join('~', '.ssh', 'id_rsa')))
    transport = paramiko.Transport((credentials.get('domain'), int(port)))
    sftp = None
    c = 0
    while c <= retry:
        try:
            transport.connect(username=credentials.get('user'), pkey=private_key)
            sftp = paramiko.SFTPClient.from_transport(transport)
            sftp.chdir(credentials.get('folder'))
            sftp.put(filename, os.path.basename(filename))
            c = retry + 1
        except paramiko.ssh_exception.SSHException:
            c += 1
            if c >= retry:
                raise
            logger.warning('Error while uploading the file, retry %s/%s', c, retry)
        finally:
            if sftp:
                sftp.close()
            if transport:
                transport.close()


def upload_file(file_name, url):
    """ Uploads a file to the desired url using the matching protocol

    :param file_name: The full or relative path to the file that you want to upload
    :param url: the url and path to upload the file to with
                the following format: protocol://[user@]domain.com[:port]/remote/path
                if no port is provided will use 22, and if no user is provided will
                use the O.S. user that is executing the command
    """
    credentials = parse_url(url)
    if credentials.get('protocol') == 'sftp':
        upload_scp(file_name, credentials)
    else:
        raise NotImplementedError('Protocol {protocol} not implemented yet'
                                  .format(protocol=credentials.get('protocol')))


def get_backup_src(config_dict=False, deployv_config=False):
    res = False
    if config_dict and "backup_folder" in config_dict.get("container_config", {}):
        res = os.path.expanduser(config_dict.get("container_config").get("backup_folder"))
    elif deployv_config and deployv_config.has_option("deployer", "backup_folder"):
        res = os.path.expanduser(deployv_config.get("deployer", "backup_folder"))
    return res


def parse_backup_size(size):
    """Parses the backup size provided and returns the value and unit.

    :param size: string with the backup size in the format `5GB`
    :return: dictionary with the key `size_value` that contains the value of the size (int)
        and the `size_unit` key containing the unit (str)
    """
    res = {}
    if not size:
        return res
    size = size.replace(' ', '').upper()
    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    exp = r'(?P<size>[\d]*([\.][\d]*)?)(?P<unit>{units})'.format(units="|".join(units))
    match = re.match(exp, size)
    if not match:
        return res
    size = float(match.group('size'))
    unit = match.group('unit')
    res.update({'size_value': size, 'size_unit': unit})
    return res


def byte_converter(from_str, to_unit='B', return_float=False):
    """Converts from one byte unit to another.

    Supported units: B, KB, MB, GB, TB, PB, EB, ZB, YB.

    :param from_str: A measure of bytes in string form. For example: '1KB', '1GB'.
    :type from_str: str
    :param to_unit: The unit you want to convert to. For example: 'GB', 'B' (default).
    :type to_unit: str
    :param return_float: If False (default) it will return the new size (always with two
        decimals) in a string form (for example: '1.00TB'). If True, it will return
        just the size as a float.
    :type return_float: bool

    :returns: The converted measure.
    :rtype: float or str
    """
    to_unit = to_unit.upper()
    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    if to_unit not in units:
        return False
    from_size = parse_backup_size(from_str)
    if not from_size or from_size['size_unit'] not in units:
        return False
    base_size = 1024
    steps = units.index(from_size['size_unit']) - units.index(to_unit)
    if steps:
        to_size = from_size['size_value'] * (base_size ** steps)
    else:
        to_size = from_size['size_value'] / (base_size ** abs(steps))
    to_str = '{size:.2f}{unit}'.format(size=to_size, unit=to_unit)
    return to_size if return_float else to_str


def verify_free_disk_space(expected_free_space, directory=False):
    """Verifies that the disk has enough available space.

    :param expected_free_space: A measure of bytes in string form.
      For example: '10GB', '5MB'.
    :type expected_free_space: str
    :param directory: Directory to verify for available space. Defaults to `/`
    :type directory: str

    :returns: Whether the disk has enough space.
    :rtype: bool
    """
    directory = directory or '/'
    stats = os.statvfs(directory)
    available = float(stats.f_frsize * stats.f_bavail)
    expected = byte_converter(expected_free_space, return_float=True)
    return available >= expected


def get_certificates_pac():
    certificate_path = deployv_static.get_template_path(
        'l10n_mx_edi.certificate.csv')
    try:
        csvfile = open(certificate_path, 'r')
        certificates = list(csv.DictReader(csvfile))[0]
    except IOError:
        certificates = {}
    return certificates


def render_template(template, values=False):
    """Renders the provided jinja template with the provided values and
    returns the rendered template.
    """
    values = values or {}
    template_path = deployv_static.get_template_path(template)
    env = Environment(loader=FileSystemLoader(os.path.dirname(template_path)))
    template = env.get_template(template)
    rendered_template = template.render(values)
    return rendered_template


def is_base64_encode(string_encode):
    """verifies that the string is a base64 encoding
    """
    if not string_encode:
        return False
    try:
        base64.standard_b64decode(string_encode)
    except (binascii.Error, TypeError):
        return False
    return True


def validate_command(message, server_type):
    if not server_type:
        return {}
    command_name = message.get('command') if message.get('command') != 'ping' else False
    validate = (command_name and _AVAILABLES.get(server_type) and
                command_name not in _AVAILABLES.get(server_type))
    if validate:
        return {'error': ("the command {} is restricted for this server"
                          .format(command_name))}
    return {}


def get_size(path):
    """Gets the size of the provided file or directory in the format `5GB`,
    if an error occurs, this will return False
    """
    shell = spur.LocalShell()
    try:
        res = shell.run(shlex.split('du -sh {path}'.format(path=path)))
    except spur.results.RunProcessError as error:
        error_msg = 'Failed to get the size of {path}: {error}'.format(
            path=path, error=get_error_message(error))
        logger.warning(error_msg)
        return False
    # Get the size only and add `B` so it has the format `5GB` instead of `5G`,
    # so the methods that parse the sizes can accept it
    res = res.output.split()[0].decode() + 'B'
    return res


def get_images_path():
    """Gets the path where the docker images are stored.
    """
    shell = spur.LocalShell()
    res = shell.run(shlex.split('docker info -f "{{.DockerRootDir}}"')).output
    return res.strip(b'\n')


def get_ip_address(ifname):
    tmp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
        tmp_sock.fileno(),
        0x8915,
        struct.pack('256s', encode(ifname[:15]))
    )[20:24])


def is_docker_ip(ip_addr, docker_ip=False):
    """Recevives an ip and checks if it is in the docker network range.
    The optional `docker_ip` parameter is used to receive the docker gateway IP
    from another source instead of obtaining it from the docker0 network. This is
    useful in cases where the docker0 network doesn't exist, e.g: when we use deployv
    inside a container.
    """
    ip_addr = socket.gethostbyname(ip_addr)
    docker_ip = docker_ip or get_ip_address('docker0')
    docker0 = re.sub(r"\.\d{1,3}$", ".0/24", docker_ip)
    return ipaddress.ip_address(decode(ip_addr)) in ipaddress.ip_network(decode(docker0))


def find_files(file_path, file_name):
    """Searches for files with the provided name in the provided path and
    return the list (empty if none is found)

    :param file_path: Path where the search will be done
    :param file_name: Name of the file to find
    :return: The contents of the file or False if the file was not found
    """
    shell = spur.LocalShell()
    file_path = shell.run(['find', file_path, '-name', file_name])
    res = []
    if file_path:
        lines = file_path.output.decode('utf8').split('\n')
        res = [l.strip() for l in lines if l.strip()]
    return res


def compare_repositories(repositories, new_repositories):
    """Compare the `new_repositories` with the `repositories`, it compare with
    the `path` where store the repositories, it returns the repositories
    that have any differentiate or the repository doesn't found in `repositories`.

    :param repositories: The repositories that will compare with `new_repositories`.
    :type: list
    :param new_repositories: The repositories that will compare with `repositories`.
    :type: list
    :return: The repositories that don't belong in `repositories` or have some diff.
    :rtype: list
    """
    repos_paths = {repo.get('path'): repo for repo in repositories}
    repos_diff = []
    exclude_keys = ["root['main_repo']", "root['depth']", "root['is_dirty']"]
    for new_repo in new_repositories:
        repo = repos_paths.pop(new_repo.get('path'), {})
        if not repo:
            repos_diff.append(new_repo)
            continue
        for key, value in new_repo.items():
            if key == 'repo_url':
                value.update({'origin': encode(value.get('origin'))})
                continue
            if key == 'commit':
                new_repo.update({key: encode(value[:7])})
                repo.update({key: repo.get(key)[:7]})
                continue
            new_repo.update({key: encode(value)})
        diff = DeepDiff(new_repo, repo, exclude_paths=exclude_keys)
        if diff.get('type_changes'):
            repos_diff.append(new_repo)
    return repos_diff


def parse_variables_to_json(variables):
    """Convert to dict the env variables used to execute a process in the instance.

    :param variables: the enviroment variables used to build/create/restore
        and other things that can do with the instance.
    :type: str
    :return: a dict with the information getted in `variables`.
    :rtype: dict
    """
    if not variables:
        return {}
    if isinstance(variables, list):
        vs = [v.replace('export', '').strip() for v in variables]
    else:
        vs = variables.replace('export', '').strip().split('\n')
    return container.parse_env_vars(vs)


def add_repo(branch, name, repo_path, repo_url, commit=False, config=False,
             overwrite_repos=False, extra=False):
    """Creates a dictionary with the info of the specified repository and
    appends it to the repositories list in the config dictionary or
    return the dict created.

    :param config: Config dict where the new repo will be added
    :param name: Name of the repository
    :param path: Path inside ~/instance where the repository will be cloned
    :param repo_url: ssh or https URL used to clone the repo
    :param branch: branch of the repo we want to clone
    :param commit: the commit of repository
    :param overwrite_repos: Identificator to overwrite the repositories.
    :type overwrite_repos: bool
    :return: Config dict with the new repo
    """
    repo = {
        'branch': branch, 'name': name, 'path': repo_path,
        'commit': commit or '', 'depth': 1,
        'repo_url': {'origin': repo_url}
    }
    repo.update(extra or {})
    if not config:
        return repo
    if overwrite_repos:
        config.get('instance', {}).update({'repositories': [repo]})
    else:
        config.get('instance', {}).get('repositories', []).append(repo)
    return config


def parse_repo_url(repo_url):
    """Parses the provided url to get its parts such as the protocol,
    the org, the repo name, and the domain. The url can be in either format:

    https://github.com/organization/repo
    git@github.com:organization/repo.git

    :param repo_url: URL to be parsed
    :type: string
    :return: dictionary with the parts of the url, if the url contains
        git@ instead of a proper protocol, https:// will be returned
        as the protocol
    """
    if not repo_url:
        return {}
    regex = (
        r'(?P<protocol>git@|http[s]?:\/\/)(?P<domain>[\.\w\d-]*)[:\/]{1}'
        r'(?P<repo>(?P<namespace>[\w\d-]*)/(?P<repo_name>[\w\d-]*))(\.git)?'
    )
    res = re.match(regex, repo_url)
    if not res:
        return {}
    values = res.groupdict()
    if values['protocol'] == 'git@':
        values.update({'protocol': 'https://'})
    return values


def read_lines(filename):
    """Returns a file content as a list, no blank lines and stipped

    :param filename: Fill path to the file
    :return: A list with the file content, empty if no lines
    """
    with open(filename, 'r') as obj:
        file_contents = obj.read()
    lines = file_contents.split('\n')
    res = [l.strip() for l in lines if l.strip()]
    return res


def deploy_key(ssh_key, folder_path):
    """Deploy a SSH private key into the temporal working folder.
    """
    if not is_base64_encode(ssh_key):
        logger.warning("Can not use the specified ssh key. It is not encoded in base64")
        return False
    logger.info('Deploying keys')
    ssh_key = decode(base64.b64decode(ssh_key))
    filename = os.path.join(folder_path, 'id_rsa')
    with open(filename, 'w') as key_file:
        key_file.write(ssh_key)
    os.chmod(filename, S_IREAD + S_IWRITE)
    return filename


def get_main_repo(repos):
    """Get the main repository from a list of repositories.
    """
    main = [r for r in repos if r.get('main_repo')]
    return main and main[0] or {}



# coding: utf-8

from os import path
from six import string_types
import six


def generate_hostname(config):
    """ Helper method that generates the hostname according to Vauxoo standard:

        container_name.domain_serving.the.instance.com

    :param config: Configuration Object

    :return: String formated as "{{prefix}}.{{domain}}"
    """
    res = "{}.{}".format(
        generate_prefix(config),
        config.container_config.get('domain'))
    return res.replace('_', '')


def generate_binds(volumes, working_folder):
    """ Helper method to generate volume binds as required by docker-py

        http://docker-py.readthedocs.org/en/latest/volumes/

        In case the working folder is prefixed with ~ it'll be expanded to the user's
        home directory

    :param volumes: dict with volumes in the following format:

        .. code-block:: python

            {
                "host_folder_name": "volume_inside_docker",
                "host_folder_name_2": "volume_inside_docker_2",
            }

    :param working_folder: base folder in the host where the volumes mounted inside the container
        will be created

    :returns: Dict with the following format:

        .. code-block:: python

            {
                "{{ working_folder }}/host_folder_name": {
                    "bind": "/volume_inside_docker",
                    "ro": False
                    },
                "{{ working_folder }}/host_folder_name_2": {
                    "bind": "/volume_inside_docker_2",
                    "ro": False
                    }
            }

    """
    volume_binds = {}
    for key, value in six.iteritems(volumes):
        paths = value.split(':')
        host_path = ''
        instance_path = ''
        if len(paths) > 1:
            host_path, instance_path = paths[:]
        volume = path.expanduser(host_path or path.join(working_folder, key))
        volume_binds.update({
            volume: {
                "bind": instance_path or value,
                "ro": False,
            }
        })
    return volume_binds


def generate_prefix(config):
    """ Helper method that generates the prefix for containers name, folders and databases
        to be created.

    :param config: Object with the deployment configuration

    :return: String formatted as "{{task_id}}_{{customer_id}}"
    """
    if config.prefix:
        return config.prefix
    task = config.instance_config.get('task_id')
    customer = config.instance_config.get('customer_id')
    return "{}_{}".format(task, customer) if task and customer else False


def generate_env_vars(config):
    """ Helper method to generate list of env vars in the format required by docker-py
        http://docker-py.readthedocs.org/en/latest/api/#create_container

    :param config: dict with env vars and values as follows:

        .. code-block:: python

            {
                "var1": "value1",
                "var2": 234,
            }

    :return: List of env vars, vars names are uppercased as most of unixlike OS ["VAR1=value1",
        "VAR2=234"]
    """
    return ["{}={}".format(key.upper(), value) for key, value in six.iteritems(config)]


def generate_port_lists(config):
    """ Helper method to generate list of exposed ports inside container
        http://docker-py.readthedocs.org/en/latest/api/#create_container

    :param config: keys are the ports inside docker and values the ports in the host:

        .. code-block:: python

            {
                "8069": 2213,
                "8072": 2214,
            }

    :return: List of integers that represents the ports to be exposed:
            [8069, 8072]
    """
    return [int(value) for value in config.keys()]


def generate_port_bindings(config):
    """Helper method that generates the dictionary of the host ports that will be
    bound to the container's exposed ports.

    :param config: Dictionary where the keys are the ports inside docker and the values
        are the ports in the host
    :return: Parsed dictionary with the correct format of the ports
    """
    res = config.copy()
    for container_port, host_port in config.items():
        if isinstance(host_port, string_types):
            parts = host_port.split(':')
            res.update({container_port: [(parts[0], int(parts[1]))]})
        elif isinstance(host_port, list):
            res.update({container_port: list()})
            for port in host_port:
                parts = port.split(':')
                res.get(container_port).append((parts[0], int(parts[1])))
    return res


def get_ports_dict(networking):
    """ Parse the networking dict gotten from docker inspect and returns a dict with the private
        ports as keys and exposed ports as values

    :param networking: Networking configuration gotten from
        :meth:`~deployv.base.dockerv.DockerV.inspect`

    :return: Mapping ports dict:

        .. code-block:: python

            {
                '8072': 32768,
                '8069': 8069
            }

    .. todo:: Fix the reference deployv.base.dockerv.DockerV.inspect

    """
    ports = networking.get('NetworkSettings').get('Ports')
    res = dict()
    if not ports:
        return res
    for private, public in ports.items():
        private_port = private.split('/')[0]
        if not public:
            continue
        res.update({private_port: []})
        for port in public:
            ip = port.get('HostIp')
            port = int(port.get('HostPort'))
            if ip == '0.0.0.0':
                res.update({private_port: port})
                break
            res.get(private_port).append('{ip}:{port}'.format(ip=ip, port=port))
    return res


def parse_env_vars(env_vars):
    """ Parses the env var list gotten from docker inspect and returns a dict with the env vars
        in lowercase as keys

    :param env_vars: list of env vars as returned by docker inspect

    :return: Dict with the lowercase env vars

    """
    res = {}
    for var in env_vars:
        name, value = var.split('=')
        value = value.strip().replace('"', '')
        res.update({name.strip().lower(): value})
    return res
